import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.common_types import _size_2_t
import math
from typing import Union


def MP_add(a, b, alpha=1):
    # a+b*alpha
    return torch.add(a, b, alpha=alpha) / torch.sqrt(1 + alpha ** 2)


def MP_add_full(a, b, alpha=1, beta=1):
    return (a * alpha + b * beta) / torch.sqrt(alpha ** 2 + beta ** 2)


def MP_mul(a, b):
    norm = (a.norm(2, dim=-1).sqrt() * b.norm(2, dim=-1).sqrt()).unsqueeze(-1)
    return (a / norm) * b


def MP_cat(a, b, dim=-1, t=0.5):
    Na, Nb = a.shape[dim], b.shape[dim]
    C = math.sqrt((Na + Nb) / ((1. - t) ** 2 + t ** 2))

    a = a * (1. - t) / math.sqrt(Na)
    b = b * t / math.sqrt(Nb)
    return C * torch.cat((a, b), dim=dim)


class MPConv2d(nn.Conv2d):
    def __init__(
            self,
            in_channels: int,
            out_channels: int,
            kernel_size: _size_2_t,
            stride: _size_2_t = 1,
            padding: Union[str, _size_2_t] = 0,
            dilation: _size_2_t = 1,
            groups: int = 1,
            bias: bool = True,
            padding_mode: str = 'zeros',
            device=None,
            dtype=None,
            eps=1e-4
    ):
        super().__init__(in_channels, out_channels, kernel_size, stride, padding, dilation, groups, bias, padding_mode, device,
                         dtype)

        self.eps = eps
        self.fan_in_sqrt = math.sqrt(in_channels // groups * math.prod(kernel_size))

    def normalize_weight(self, weight, eps=1e-4):
        weight_shape = weight.shape
        weight, ps = weight.flatten(1)
        normed_weight = F.normalize(weight, dim=-1, eps=eps) * self.fan_in_sqrt
        return normed_weight.view(weight_shape)

    def forward(self, x):
        if self.training:
            with torch.no_grad():
                normed_weight = self.normalize_weight(self.weight, eps=self.eps)
                self.weight.copy_(normed_weight)
            weight = self.weight / self.fan_in_sqrt
        else:
            weight = self.normalize_weight(self.weight, eps=self.eps) / self.fan_in_sqrt

        output = self._conv_forward(x, weight, self.bias)
        if self.bias is None:
            return output
        else:
            return output / torch.sqrt(1 + self.bias.norm(2).square())


class Linear(nn.Linear):
    def __init__(self, in_features: int, out_features: int, bias: bool = True,
                 device=None, dtype=None, eps=1e-4):
        super().__init__(in_features, out_features, bias, device, dtype)

        self.eps = eps
        self.fan_in_sqrt = math.sqrt(in_features)

    def normalize_weight(self, weight, eps=1e-4):
        weight_shape = weight.shape
        weight, ps = weight.flatten(1)
        normed_weight = F.normalize(weight, dim=-1, eps=eps) * self.fan_in_sqrt
        return normed_weight.view(weight_shape)

    def forward(self, x):
        if self.training:
            with torch.no_grad():
                normed_weight = self.normalize_weight(self.weight, eps=self.eps)
                self.weight.copy_(normed_weight)
            weight = self.weight / self.fan_in_sqrt
        else:
            weight = self.normalize_weight(self.weight, eps=self.eps) / self.fan_in_sqrt

        output = F.linear(x, weight, self.bias)
        if self.bias is None:
            return output
        else:
            return output / torch.sqrt(1 + self.bias.norm(2).square())


class MPSiLU(nn.Module):
    def forward(self, x):
        return F.silu(x) / 0.596


class MPGELU(nn.Module):
    def forward(self, x):
        return F.gelu(x) / 0.64958


class MPGEGLU(nn.Module):
    def forward(self, x):
        x1, x2 = x.chunk(2, dim=1)
        return x1 * F.gelu(x2)
        return F.gelu(x) / 0.64958



from typing import Any, Dict, Tuple

from PIL import Image

from .base import VisionDataSource


class IndexSource(VisionDataSource):
    def __init__(self, data, repeat=1, **kwargs):
        super().__init__('', repeat=repeat)
        self.data = data

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index) -> Dict[str, Any]:
        image, label = self.data[index]
        return {
            'id': index,
            'image': image,
            'label': label
        }

    def get_image_size(self, data: Dict[str, Any]) -> Tuple[int, int]:
        image = data['image']
        if isinstance(image, Image.Image):
            return image.size
        else:
            return image.shape[1], image.shape[0]



from typing import Dict

import torch
from torch import nn

from .base import BaseWrapper
from rainbowneko.utils import maybe_DDP


class DistillationWrapper(BaseWrapper):
    def __init__(self, model_teacher, model_student, ema=None, key_map_in_student=None, key_map_in_teacher=None, key_map_out=None):
        super().__init__()
        self.model_teacher = model_teacher
        self.model_student = model_student

        self.key_map_in_student = self.build_mapper(key_map_in_student, default=('image -> 0',))
        self.key_map_in_teacher = self.build_mapper(key_map_in_teacher, default=('image -> 0',))
        self.key_mapper_out = self.build_mapper(key_map_out, default=('pred_student -> pred', 'pred_teacher -> pred_teacher'))

        if ema is not None:
            self.ema_teacher = ema(self.model_teacher)

    def forward(self, ds_name, plugin_input={}, **kwargs):
        inputs_T_args, inputs_T_kwargs = self.get_inputs_feed(self.key_map_in_teacher, self.model_teacher, kwargs, plugin_input,
                                                              ds_name=ds_name)
        inputs_S_args, inputs_S_kwargs = self.get_inputs_feed(self.key_map_in_student, maybe_DDP(self.model_student), kwargs, plugin_input,
                                                              ds_name=ds_name)

        res = {}
        if len(inputs_T_args)>0 or len(inputs_T_kwargs)>0:
            with torch.no_grad():
                if hasattr(self, 'ema_teacher'):
                    out_teacher = self.ema_teacher(*inputs_T_args, **inputs_T_kwargs)
                else:
                    out_teacher = self.model_teacher(*inputs_T_args, **inputs_T_kwargs)
            res['pred_teacher'] = out_teacher
        out_student = self.model_student(*inputs_S_args, **inputs_S_kwargs)
        res['pred_student'] = out_student

        return self.get_map_data(self.key_mapper_out, res, ds_name=ds_name)[1]

    def update_model(self, step:int):
        if hasattr(self, 'ema_teacher'):
            self.ema_teacher.step(maybe_DDP(self.model_student))

    @property
    def trainable_models(self) -> Dict[str, nn.Module]:
        return {'model_student': self.model_student}



from typing import Dict, List
from rainbowneko.ckpt_manager.locator import get_match_layers
from torch import nn

class CfgModelParser:
    def __init__(self, cfg_model: List[Dict], lr=1e-5, weight_decay=0):
        self.cfg_model = cfg_model
        self.lr = lr
        self.weight_decay = weight_decay

    def get_params(self, layers, named_modules):
        params = []
        train_layers = []
        for layer_name in get_match_layers(layers, named_modules):
            layer = named_modules[layer_name]
            layer.requires_grad_(True)
            layer.train()
            train_layers.append(layer)
            params.extend(layer.parameters())
        return train_layers, list(dict.fromkeys(params))  # remove duplicates and keep order

    def get_params_group(self, model):
        named_modules = {k: v for k, v in model.named_modules()}

        params_group = []
        train_layers = []

        if self.cfg_model is not None:
            for item in self.cfg_model:
                layers, params = self.get_params(item.layers, named_modules)
                params_group.append({"params": params, "lr": getattr(item, "lr", self.lr),
                                    "weight_decay": getattr(item, "weight_decay", self.weight_decay)})
                train_layers.extend(layers)

        return params_group, train_layers

class CfgWDModelParser(CfgModelParser):
    def get_params_group(self, model):
        named_modules = {k: v for k, v in model.named_modules()}

        params_group = []
        train_layers = []

        wd_params = set()
        for m in model.modules():
            if isinstance(m, nn.Linear) or isinstance(m, nn.Conv2d):
                wd_params.add(m.weight)

        if self.cfg_model is not None:
            for item in self.cfg_model:
                layers, params = self.get_params(item.layers, named_modules)
                train_layers.extend(layers)

                params_nowd = []
                params_wd = []
                for p in params:
                    if p in wd_params:
                        params_wd.append(p)
                    else:
                        params_nowd.append(p)
                if len(params_nowd)>0:
                    params_group.append({"params": params_nowd, "lr": getattr(item, "lr", self.lr),
                                        "weight_decay": 0})
                if len(params_wd) > 0:
                    params_group.append({"params": params_wd, "lr": getattr(item, "lr", self.lr),
                                         "weight_decay": getattr(item, "weight_decay", self.weight_decay)})

        return params_group, train_layers

class CustomModelParser:
    def __init__(self, params_loader, **kwargs):
        self.params_loader = params_loader
        self.kwargs = kwargs

    def get_params_group(self, model):
        train_params, train_layers = self.params_loader(model, **self.kwargs)
        return train_params, train_layers


from .cfg_resolvers import *
from .cfg_tool import change_num_classes
from .constant import Path_Like
from .key_mapper import KeyMapper
from .lr_scheduler import ConstantLR, CosineLR, CosineRestartLR, MultiStepLR, PolynomialLR, OneCycleLR, get_scheduler, \
    get_scheduler_with_name, SchedulerType
from .module import disable_hf_loggers
from .net import remove_layers, remove_all_hooks, to_cpu, to_cuda, hook_compile, split_module_name, xformers_available, maybe_DDP
from .random import RandomContext
from .utils import *



"""
pair_dataset.py
====================
    :Name:        text-image pair dataset
    :Author:      Dong Ziyi
    :Affiliation: HCP Lab, SYSU
    :Created:     10/03/2023
    :Licence:     Apache-2.0
"""

from typing import Dict, Tuple, Union, List, Callable

import torch
import numpy as np
from torch.utils.data import Dataset

from .bucket import BaseBucket
from .source import DataSource, ComposeDataSource
from rainbowneko.utils import RandomContext
from .handler import DataHandler

class BaseDataset(Dataset):
    """
    A dataset to prepare the instance and class images with the prompts for fine-tuning the model.
    It pre-processes the images and the tokenizes prompts.
    """

    def __init__(self, bucket: BaseBucket = None, source: Dict[str, DataSource] = None, handler: DataHandler = None,
                 batch_handler: DataHandler=None, **kwargs):
        self.bucket: BaseBucket = bucket
        self.source = ComposeDataSource(list(source.values()))
        self.handler = handler
        self.batch_handler = batch_handler

    def build_bucket(self, bs, world_size):
        self.bucket.build(bs=bs, world_size=world_size, source=self.source)

    def batch_process(self, batch: Dict[str, Union[List, torch.Tensor]]):
        if self.batch_handler is not None:
            batch = self.batch_handler(batch)
        return batch

    def __len__(self):
        return len(self.bucket)

    def __getitem__(self, index):
        datas = self.bucket[index]
        datas = self.handler(datas)
        return datas

    def collate_fn(self, batch):
        '''
        batch: [{
            img:tensor,
            label:tensor,
            ...,
            plugin_input:{...}
        }]
        '''
        datas = {k: [] for k in batch[0].keys()}

        has_plugin_input = 'plugin_input' in batch[0]
        if has_plugin_input:
            plugin_input = {k: [] for k in batch[0]['plugin_input'].keys()}
            del datas['plugin_input']

        for data in batch:
            if has_plugin_input:
                for k, v in data.pop('plugin_input').items():
                    plugin_input[k].append(v)

            for k, v in data.items():
                datas[k].append(v)

        def batch_merge(data):
            for k, v in data.items():
                if isinstance(v[0], torch.Tensor):
                    data[k] = torch.stack(v)
                elif isinstance(v[0], dict):
                    pass
                else:
                    try:
                        data[k] = self.create_tensor(v)
                    except:
                        pass
            return data

        datas = batch_merge(datas)

        if has_plugin_input:
            datas['plugin_input'] = batch_merge(plugin_input)

        datas = self.batch_process(datas)

        return datas

    @staticmethod
    def create_tensor(data):
        if isinstance(data, list) and isinstance(data[0], np.ndarray):
            return torch.tensor(np.array(data))
        else:
            return torch.tensor(data)



import argparse
import re

import yaml
from omegaconf import OmegaConf
from omegaconf.omegaconf import get_omega_conf_dumper, _ensure_container
from rainbowneko.parser import PythonCfgParser


class NoAliasDumper(get_omega_conf_dumper()):
    def ignore_aliases(self, data):
        return True


def to_yaml(cfg, *, resolve: bool = False, sort_keys: bool = False, alias: bool = False) -> str:
    """
    returns a yaml dump of this config object.

    :param cfg: Config object, Structured Config type or instance
    :param resolve: if True, will return a string with the interpolations resolved, otherwise
        interpolations are preserved
    :param sort_keys: If True, will print dict keys in sorted order. default False.
    :return: A string containing the yaml representation.
    """
    cfg = _ensure_container(cfg)
    container = OmegaConf.to_container(cfg, resolve=resolve, enum_to_str=True)
    return yaml.dump(  # type: ignore
        container,
        default_flow_style=False,
        allow_unicode=True,
        sort_keys=sort_keys,
        Dumper=get_omega_conf_dumper() if alias else NoAliasDumper,
    )


def remove_wrapping(text, prefix='!!python/name:', suffix=" ''"):
    pattern = prefix + r"([^'\n]+)" + suffix
    result = re.sub(pattern, r'\1', text)
    return result


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Python cfg to yaml cfg")
    parser.add_argument("cfg", type=str, default=None)
    parser.add_argument('--alias', action='store_true')
    parser.add_argument('--save', type=str, default=None)
    args = parser.parse_args()

    parser = PythonCfgParser()
    py_cfg = parser.load_cfg(args.cfg)
    yaml_cfg = to_yaml(py_cfg)
    yaml_cfg = remove_wrapping(yaml_cfg)
    #yaml_cfg = remove_wrapping(yaml_cfg, prefix='!!python/object:', suffix='')

    if args.save is None:
        print(yaml_cfg)
    else:
        with open(args.save, 'w') as f:
            f.write(yaml_cfg)



from typing import Dict, Union, List, Callable

import torch
from rainbowneko.utils import Path_Like
from contextlib import contextmanager

from .dataset import BaseDataset, BaseBucket, DataSource, DataHandler


class DataCache:
    def __init__(self, pre_build: Path_Like = None):
        self.pre_build = pre_build
        self.cache = self.load(pre_build) if pre_build else {}

    def before_load(self, index):
        return None, False

    def before_handler(self, index, data):
        return data

    def on_finish(self, index, data):
        if index in self.cache:
            return self.cache[index]
        else:
            self.cache[index] = data
            return data

    def on_batch(self, batch):
        return batch

    def build(self, dataset, model, all_gather: Callable):
        self.cache.clear()

    def load(self, path):
        return torch.load(path)


class DataCacheGroup(DataCache):
    def __init__(self, *caches):
        self.caches = caches

    def before_load(self, index):
        new_data = {}
        for cache in self.caches:
            data, full = cache.before_load(index)
            if full:
                raise ValueError('caches in DataCacheGroup should not be full cache!')
            new_data.update(data)
        return new_data, False


class CacheableDataset(BaseDataset):
    def __init__(self, bucket: BaseBucket = None, source: Dict[str, DataSource] = None, handler: DataHandler = None,
                 batch_handler: DataHandler = None, cache: DataCache = None, **kwargs):
        super().__init__(bucket=bucket, source=source, handler=handler, batch_handler=batch_handler, **kwargs)
        self.cache = cache
        self._disable = False

    def __getitem__(self, index):
        if self._disable or self.cache is None:
            return super().__getitem__(index)
        datas, full = self.cache.before_load(index)
        if full:
            return datas
        datas = datas or self.bucket[index]
        datas = self.cache.before_handler(index, datas)
        datas = self.handler(datas)
        datas = self.cache.on_finish(index, datas)
        return datas

    @contextmanager
    def disable_cache(self):
        self._disable = True
        yield
        self._disable = False

    def batch_process(self, batch: Dict[str, Union[List, torch.Tensor]]):
        if self.batch_handler is not None:
            batch = self.batch_handler(batch)
        if self.cache is not None:
            batch = self.cache.on_batch(batch)
        return batch

    def build_bucket(self, bs, world_size):
        self.bucket.build(bs=bs, world_size=world_size, source=self.source)

    def build_cache(self, model, all_gather: Callable):
        if self.cache is not None:
            self.cache.build(self, model, all_gather)



from typing import Dict, Any, Union

import torch
from torch.serialization import FILE_LIKE
from .base import CkptFormat


class PKLFormat(CkptFormat):
    EXT = 'ckpt'

    def save_ckpt(self, sd_model: Union[torch.nn.Module, Dict[str, Any]], save_f: FILE_LIKE):
        torch.save(sd_model, save_f)

    def load_ckpt(self, ckpt_f: FILE_LIKE, map_location="cpu"):
        return torch.load(ckpt_f, map_location=map_location)


from .base import SingleWrapper


class FeatWrapper(SingleWrapper):

    def forward(self, x, plugin_input={}, **kwargs):
        input_all = dict(x=x, **plugin_input)

        if hasattr(self.model, 'input_feeder'):
            for feeder in self.model.input_feeder:
                feeder(input_all)

        out, feat = self.model(x, **kwargs)
        if not isinstance(feat, list):
            feat = [feat]
        return {'pred': out, 'feat': feat}



import math
import os.path
import pickle
from concurrent.futures import ThreadPoolExecutor
from typing import Callable

import numpy as np
from loguru import logger
from sklearn.cluster import KMeans
from tqdm import tqdm
from rainbowneko.utils import repeat_list

from .base import BaseBucket
from ..handler import AutoSizeHandler


class RatioBucket(BaseBucket):
    can_shuffle = False
    handler = AutoSizeHandler()

    def __init__(self, target_area: int = 640 * 640, step_size: int = 8, num_bucket: int = 10, pre_build_bucket: str = None):
        self.target_area = target_area
        self.step_size = step_size
        self.num_bucket = num_bucket
        self.pre_build_bucket = pre_build_bucket

    def load_bucket(self, path):
        with open(path, 'rb') as f:
            data = pickle.load(f)
        self.buckets = data['buckets']
        self.size_buckets = data['size_buckets']
        self.idx_bucket_map = data['idx_bucket_map']
        self.data_len = data['data_len']

    def save_bucket(self, path):
        with open(path, 'wb') as f:
            pickle.dump({
                'buckets': self.buckets,
                'size_buckets': self.size_buckets,
                'idx_bucket_map': self.idx_bucket_map,
                'data_len': self.data_len,
            }, f)

    def build_buckets_from_ratios(self):
        logger.info('build buckets from ratios')
        size_low = int(math.sqrt(self.target_area / self.ratio_max))
        size_high = int(self.ratio_max * size_low)

        # SD需要边长是8的倍数
        size_low = (size_low // self.step_size) * self.step_size
        size_high = (size_high // self.step_size) * self.step_size

        data = []
        for w in range(size_low, size_high + 1, self.step_size):
            for h in range(size_low, size_high + 1, self.step_size):
                data.append([w * h, np.log2(w / h), w, h])  # 对比例取对数，更符合人感知，宽高相反的可以对称分布。
        data = np.array(data)

        error_area = np.abs(data[:, 0] - self.target_area)
        data_use = data[np.argsort(error_area)[:self.num_bucket * 3], :]  # 取最小的num_bucket*3个

        # 聚类，选出指定个数的bucket
        kmeans = KMeans(n_clusters=self.num_bucket, random_state=114514).fit(data_use[:, 1].reshape(-1, 1))
        labels = kmeans.labels_
        self.buckets = []  # [bucket_id:[file_idx,...]]
        ratios_log = []
        self.size_buckets = []
        for i in range(self.num_bucket):
            map_idx = np.where(labels == i)[0]
            m_idx = map_idx[np.argmin(np.abs(data_use[labels == i, 1] - np.median(data_use[labels == i, 1])))]
            # self.buckets[wh_hash(*data_use[m_idx, 2:])]=[]
            self.buckets.append([])
            ratios_log.append(data_use[m_idx, 1])
            self.size_buckets.append(data_use[m_idx, 2:].astype(int))
        ratios_log = np.array(ratios_log)
        self.size_buckets = np.array(self.size_buckets)

        if self.source_indexable:
            # get images ratio
            def get_ratio(info):
                data, source = info
                w, h = source.get_image_size(data)
                ratio = np.log2(w / h)
                return ratio

            ratio_list = []
            with self.source.return_source(), ThreadPoolExecutor() as executor:
                for ratio in tqdm(executor.map(get_ratio, self.source), desc='get image info', total=len(self.source)):
                    ratio_list.append(ratio)
            ratio_list = np.array(ratio_list)

            # fill buckets with images w,h
            bucket_id = np.abs(ratio_list[:, None] - ratios_log[None, :]).argmin(axis=-1)
            self.idx_bucket_map = bucket_id
            for bidx in range(self.num_bucket):
                self.buckets.append(np.where(bucket_id == bidx)[0].tolist())
            logger.info(
                'buckets info: ' + ', '.join(f'size:{self.size_buckets[i]}, num:{len(b)}' for i, b in enumerate(self.buckets)))
        else:
            self.ratios_log = ratios_log
            logger.info('buckets info: ' + ', '.join(f'size:{size}' for size in enumerate(self.size_buckets)))

    def build_buckets_from_images(self):
        logger.info('build buckets from images')

        def get_ratio(info):
            data, source = info
            w, h = source.get_image_size(data)
            ratio = np.log2(w / h)
            return ratio

        ratio_list = []
        with self.source.return_source(), ThreadPoolExecutor() as executor:
            for ratio in tqdm(executor.map(get_ratio, self.source), desc='get image info', total=len(self.source)):
                ratio_list.append(ratio)
        ratio_list = np.array(ratio_list)

        # 聚类，选出指定个数的bucket
        kmeans = KMeans(n_clusters=self.num_bucket, random_state=114514, verbose=True, tol=1e-3).fit(ratio_list.reshape(-1, 1))
        labels = kmeans.labels_
        ratios = 2 ** kmeans.cluster_centers_.reshape(-1)

        h_all = np.sqrt(self.target_area / ratios)
        w_all = h_all * ratios

        # SD需要边长是8的倍数
        h_all = (np.round(h_all / self.step_size) * self.step_size).astype(int)
        w_all = (np.round(w_all / self.step_size) * self.step_size).astype(int)
        self.size_buckets = list(zip(w_all, h_all))
        self.size_buckets = np.array(self.size_buckets)

        if self.source_indexable:
            self.buckets = []  # [bucket_id:[file_idx,...]]
            self.idx_bucket_map = labels
            for bidx in range(self.num_bucket):
                self.buckets.append(np.where(labels == bidx)[0].tolist())
            logger.info(
                'buckets info: ' + ', '.join(f'size:{self.size_buckets[i]}, num:{len(b)}' for i, b in enumerate(self.buckets)))
        else:
            self.ratios_log = np.log2(ratios)
            logger.info('buckets info: ' + ', '.join(f'size:{size}' for size in enumerate(self.size_buckets)))

    def build(self, bs: int, world_size: int, source: 'ComposeDataSource'):
        '''
        :param bs: batch_size * n_gpus * accumulation_step
        :param img_root_list:
        '''
        self.source = source
        self.bs = bs * world_size        

        try:
            _ = self.source[0]
            self.source_indexable = True
        except NotImplementedError:
            self.source_indexable = False

        if self.pre_build_bucket and os.path.exists(self.pre_build_bucket):
            self.load_bucket(self.pre_build_bucket)
            return

        self._build()

        rs = np.random.RandomState(42)
        # make len(bucket)%bs==0
        self.data_len = 0
        for bidx, bucket in enumerate(self.buckets):
            rest = len(bucket) % self.bs
            if rest > 0:
                bucket.extend(rs.choice(bucket, self.bs - rest))
            self.data_len += len(bucket)
            self.buckets[bidx] = np.array(bucket)

        if self.pre_build_bucket:
            self.save_bucket(self.pre_build_bucket)

    def rest(self, epoch):
        if self.source_indexable:
            rs = np.random.RandomState(42 + epoch)
            bucket_list = [x.copy() for x in self.buckets]
            # shuffle inter bucket
            for x in bucket_list:
                rs.shuffle(x)

            # shuffle of batches
            bucket_list = np.hstack(bucket_list).reshape(-1, self.bs).astype(int)
            rs.shuffle(bucket_list)

            self.idx_bucket = bucket_list.reshape(-1)
        else:
            self.rs = np.random.RandomState(42 + epoch)

    def _buffer(self, bs, assign_bucket:Callable, bufsize=1000, initial=100, rs:np.random.RandomState=None):
        """Bucket and shuffle the data in the stream.

        This uses a buffer of size `bufsize`. Bucketing and shuffling datas of non-indexable source.

        bs: batch size
        assign_bucket: function for assign data to bucket
        bufsize: buffer size
        returns: iterator
        """
        initial = min(initial, bufsize)
        buckets = [[] for _ in self.size_buckets]
        count = 0

        source_iter = iter(self.source)

        select_bucket = None
        bs_count=bs

        def pick():
            nonlocal count
            if rs is None:
                k = 0
            else:
                k = rs.randint(0, len(select_bucket) - 1)
            sample = select_bucket[k]
            select_bucket[k] = select_bucket[-1]
            select_bucket.pop()
            count -= 1
            return sample

        with self.source.return_source():
            for datas, source in source_iter:
                assign_bucket(datas, source, buckets)
                count += 1

                if count < bufsize:
                    try:
                        assign_bucket(*next(source_iter), buckets)
                        count += 1
                    except StopIteration:
                        pass

                if count >= initial:
                    if bs_count>=bs:
                        candidate_bucket = [b for b in buckets if len(b) >= bs]
                        if len(candidate_bucket) == 0:
                            continue
                        if rs is None:
                            select_bucket = candidate_bucket[0]
                        else:
                            select_bucket = rs.choice(candidate_bucket)
                        bs_count = 0

                    yield pick()

            while count > 0:
                if bs_count >= bs:
                    candidate_bucket = [b for b in buckets if len(b) >= bs]
                    if len(candidate_bucket) == 0:
                        for b in buckets:
                            rest = bs - len(b)
                            if rs is None:
                                b.extend(repeat_list(b, rest))
                            else:
                                b.extend(rs.choice(b, rest))
                            count += rest
                    if rs is None:
                        select_bucket = candidate_bucket[0]
                    else:
                        select_bucket = rs.choice(candidate_bucket)
                    bs_count = 0

                yield pick()

    def next_data(self, shuffle=True):
        def assign_bucket(datas, source, buckets):
            w, h = source.get_image_size(datas)
            ratio_log = np.log2(w / h)
            bucket_idx = (self.ratios_log - ratio_log).abs().argmin()
            datas['image_size'] = self.size_buckets[bucket_idx]
            buckets[bucket_idx].append(datas)

        if not hasattr(self, 'buffer_iter'):
            self.buffer_iter = self._buffer(self.bs, assign_bucket, rs=self.rs if shuffle else None)
        return next(self.buffer_iter)

    def __getitem__(self, idx):
        if self.source_indexable:
            file_idx = self.idx_bucket[idx]
            bucket_idx = self.idx_bucket_map[file_idx]
            datas = self.source[file_idx]
            datas['image_size'] = self.size_buckets[bucket_idx]
            return datas
        else:
            return self.next_data()

    def __len__(self):
        return self.data_len

    @classmethod
    def from_ratios(cls, target_area: int = 640 * 640, step_size: int = 8, num_bucket: int = 10, ratio_max: float = 4,
                    pre_build_bucket: str = None, **kwargs):
        arb = cls(target_area, step_size, num_bucket, pre_build_bucket=pre_build_bucket)
        arb.ratio_max = ratio_max
        arb._build = arb.build_buckets_from_ratios
        return arb

    @classmethod
    def from_files(cls, target_area: int = 640 * 640, step_size: int = 8, num_bucket: int = 10, pre_build_bucket: str = None,
                   **kwargs):
        arb = cls(target_area, step_size, num_bucket, pre_build_bucket=pre_build_bucket)
        arb._build = arb.build_buckets_from_images
        return arb


class SizeBucket(RatioBucket):
    handler = AutoSizeHandler(mode='pad')

    def __init__(self, step_size: int = 8, num_bucket: int = 10, pre_build_bucket: str = None):
        super().__init__(step_size=step_size, num_bucket=num_bucket, pre_build_bucket=pre_build_bucket)

    def build_buckets_from_images(self):
        '''
        根据图像尺寸聚类，不会resize图像，只有剪裁和填充操作。
        '''
        logger.info('build buckets from images size')

        def get_size(info):
            data, source = info
            w, h = source.get_image_size(data)
            return w, h

        size_list = []
        with self.source.return_source(), ThreadPoolExecutor() as executor:
            for w, h in tqdm(executor.map(get_size, self.source), desc='get image info', total=len(self.source)):
                size_list.append([w, h])
        size_list = np.array(size_list)

        # 聚类，选出指定个数的bucket
        kmeans = KMeans(n_clusters=self.num_bucket, random_state=114514).fit(size_list)
        labels = kmeans.labels_
        size_buckets = kmeans.cluster_centers_

        # SD需要边长是8的倍数
        self.size_buckets = (np.round(size_buckets / self.step_size) * self.step_size).astype(int)

        if self.source_indexable:
            self.buckets = []  # [bucket_id:[file_idx,...]]
            self.idx_bucket_map = labels
            for bidx in range(self.num_bucket):
                self.buckets.append(np.where(labels == bidx)[0].tolist())
            logger.info(
                'buckets info: ' + ', '.join(f'size:{self.size_buckets[i]}, num:{len(b)}' for i, b in enumerate(self.buckets)))
        else:
            logger.info('buckets info: ' + ', '.join(f'size:{size}' for size in enumerate(self.size_buckets)))

    def next_data(self, shuffle=True):
        def assign_bucket(datas, source, buckets):
            w, h = source.get_image_size(datas)
            bucket_idx = np.linalg.norm(self.size_buckets-np.array([[w,h]]), axis=1).argmin()
            datas['image_size'] = self.size_buckets[bucket_idx]
            buckets[bucket_idx].append(datas)

        if not hasattr(self, 'buffer_iter'):
            self.buffer_iter = self._buffer(self.bs, assign_bucket, rs=self.rs if shuffle else None)
        return next(self.buffer_iter)

    @classmethod
    def from_files(cls, step_size: int = 8, num_bucket: int = 10, pre_build_bucket: str = None, **kwargs):
        arb = cls(step_size, num_bucket, pre_build_bucket=pre_build_bucket)
        arb._build = arb.build_buckets_from_images
        return arb


class RatioSizeBucket(RatioBucket):
    def __init__(self, step_size: int = 8, num_bucket: int = 10, max_area: int = 640 * 640, pre_build_bucket: str = None):
        super().__init__(step_size=step_size, num_bucket=num_bucket, pre_build_bucket=pre_build_bucket)
        self.max_area = max_area

    def build_buckets_from_images(self):
        '''
        根据图像尺寸聚类，不会resize图像，只有剪裁和填充操作。
        '''
        logger.info('build buckets from images')

        def get_ratio(info):
            data, source = info
            w, h = source.get_image_size(data)
            ratio = np.log2(w / h)
            log_area = 2*np.log2(min(w * h, self.max_area)/(self.max_area/2))
            return ratio, log_area

        ratio_list = []
        with self.source.return_source(), ThreadPoolExecutor() as executor:
            for ratio, log_area in tqdm(executor.map(get_ratio, self.source), desc='get image info', total=len(self.source)):
                ratio_list.append([ratio, log_area])
        ratio_list = np.array(ratio_list)

        # 聚类，选出指定个数的bucket
        kmeans = KMeans(n_clusters=self.num_bucket, random_state=114514).fit(ratio_list)
        labels = kmeans.labels_
        ratios = 2 ** kmeans.cluster_centers_[:, 0]
        sizes = 2 ** kmeans.cluster_centers_[:, 1]

        h_all = np.sqrt(sizes / ratios)
        w_all = h_all * ratios

        # SD需要边长是8的倍数
        h_all = (np.round(h_all / self.step_size) * self.step_size).astype(int)
        w_all = (np.round(w_all / self.step_size) * self.step_size).astype(int)
        self.size_buckets = list(zip(w_all, h_all))
        self.size_buckets = np.array(self.size_buckets)

        if self.source_indexable:
            self.buckets = []  # [bucket_id:[file_idx,...]]
            self.idx_bucket_map = labels
            for bidx in range(self.num_bucket):
                self.buckets.append(np.where(labels == bidx)[0].tolist())
            logger.info(
                'buckets info: ' + ', '.join(f'size:{self.size_buckets[i]}, num:{len(b)}' for i, b in enumerate(self.buckets)))
        else:
            logger.info('buckets info: ' + ', '.join(f'size:{size}' for size in enumerate(self.size_buckets)))
            self.ratio_area_list = ratio_list

    def next_data(self, shuffle=True):
        def assign_bucket(datas, source, buckets):
            w, h = source.get_image_size(datas)
            log_ratio = np.log2(w / h)
            log_area = 2 * np.log2(min(w * h, self.max_area) / (self.max_area / 2))
            bucket_idx = np.linalg.norm(self.ratio_area_list-np.array([[log_ratio,log_area]]), axis=1).argmin()
            datas['image_size'] = self.size_buckets[bucket_idx]
            buckets[bucket_idx].append(datas)

        if not hasattr(self, 'buffer_iter'):
            self.buffer_iter = self._buffer(self.bs, assign_bucket, rs=self.rs if shuffle else None)
        return next(self.buffer_iter)

    @classmethod
    def from_files(cls, step_size: int = 8, num_bucket: int = 10, max_area: int = 640 * 640, pre_build_bucket: str = None,
                   **kwargs):
        arb = cls(step_size, num_bucket, max_area=max_area, pre_build_bucket=pre_build_bucket)
        arb._build = arb.build_buckets_from_images
        return arb



from typing import Dict, Any

from torch import nn

from rainbowneko.models.plugin import PluginGroup
from .format import CkptFormat, SafeTensorFormat
from .source import LocalCkptSource

LAYERS_ALL = 'all'
LAYERS_TRAINABLE = 'trainable'


class NekoLoader:
    def __init__(self, format: CkptFormat = None, source: LocalCkptSource = None, layers='all'):
        if format is None:
            format = SafeTensorFormat()
        if source is None:
            source = LocalCkptSource()

        self.format = format
        self.source = source
        self.layers = layers

    def load(self, path, ext=None, **kwargs):
        return self.source.get(path, self.format, **kwargs)

    def load_to(self, name, model):
        raise NotImplementedError()

    @staticmethod
    def load_all(model: nn.Module, cfg: Dict[str, "NekoLoader"]):
        for name, loader in cfg.items():
            loader.load_to(name, model)


class NekoSaver:
    def __init__(self, format: CkptFormat = None, source: LocalCkptSource = None, layers='all', state_prefix=''):
        if format is None:
            format = SafeTensorFormat()
        if source is None:
            source = LocalCkptSource()
        self.format = format
        self.source = source
        self.layers = layers
        self.state_prefix = state_prefix

    def clean_prefix(self, state_dict: Dict[str, Any]):
        return {k.removeprefix(self.state_prefix): v for k, v in state_dict.items() if k.startswith(self.state_prefix)}

    def save(self, state_dict: Dict[str, Any], name, prefix=None):
        self.source.put(f"{name}.{self.format.EXT}", state_dict, self.format, prefix=prefix)

    def save_to(self, name, model: nn.Module, plugin_groups: Dict[str, PluginGroup], model_ema=None, exclude_key=None,
                name_template=None):
        raise NotImplementedError

    @staticmethod
    def save_all(model: nn.Module, plugin_groups: Dict[str, PluginGroup], cfg: Dict[str, "NekoSaver"], model_ema=None,
                 exclude_key=None, name_template=None):
        for name, loader in cfg.items():
            loader.save_to(name, model, plugin_groups, model_ema=model_ema, exclude_key=exclude_key, name_template=name_template)



from .base import BasicAction
from typing import Union, Dict
# from rainbowneko.evaluate import BaseMetric

class MetricAction(BasicAction):
    def __init__(self, metric: Union["BaseMetric", Dict[str, "BaseMetric"]], key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.metric = metric

    def forward(self, pred, target, device, _metric=None, **states):
        if _metric is None:
            self.metric.to(device)
            self.metric.reset()
        self.metric.update(pred, target)
        return {'_metric':self.metric}

# class MetricFinishAction(BasicAction):
#     def __init__(self, key_map_in=None, key_map_out=None):
#         super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
#
#     def forward(self, _metric, **states):
#         v_metric = _metric.finish(self.trainer.accelerator.gather, self.trainer.is_local_main_process)
#         return {'v_metric':_metric.finish()}


import os
from typing import Dict, Any

import albumentations as A
import cv2
import numpy as np
import torch
from PIL import Image

from .base import DataHandler
from ..utils import resize_crop_fix, pad_crop_fix
from rainbowneko.utils import Path_Like


class LoadImageHandler(DataHandler):
    def __init__(self, bg_color=(255, 255, 255), mode='RGB', key_map_in=('image -> image',), key_map_out=('image -> image',)):
        super().__init__(key_map_in, key_map_out)
        self.bg_color = bg_color
        self.mode = mode

    def load_image(self, path) -> Image.Image:
        image = Image.open(path)
        if image.mode == 'RGBA':
            x, y = image.size
            canvas = Image.new('RGBA', image.size, self.bg_color)
            canvas.paste(image, (0, 0, x, y), image)
            image = canvas
        return image.convert(self.mode)

    def handle(self, image):
        if isinstance(image, Path_Like):
            image = self.load_image(image)
        elif isinstance(image, Image.Image):
            image = image
        elif isinstance(image, np.ndarray):
            image = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
        else:
            raise NotImplementedError(f'image with type {type(image)} not supported')
        return {'image': image}

class ImageHandler(DataHandler):
    def __init__(self, transform, bg_color=(255, 255, 255), key_map_in=('image -> image',), key_map_out=('image -> image',)):
        super().__init__(key_map_in, key_map_out)
        self.transform = transform
        self.bg_color = bg_color


    def load_image(self, path) -> Dict[str, Any]:
        path = os.path.join(path)
        image = Image.open(path)
        if image.mode == 'RGBA':
            x, y = image.size
            canvas = Image.new('RGBA', image.size, self.bg_color)
            canvas.paste(image, (0, 0, x, y), image)
            image = canvas
        return image.convert("RGB")

    def procees_image(self, image):
        if isinstance(self.transform, (A.BaseCompose, A.BasicTransform)):
            image_A = self.transform(image=np.array(image))
            if isinstance(image_A['image'], np.ndarray):
                image = Image.fromarray(image_A['image'])
            else:
                image = image_A['image']
        else:
            image = self.transform(image)
        return image

    def handle(self, image):
        if isinstance(image, str):
            image = self.load_image(image)
        elif isinstance(image, (Image.Image, torch.Tensor)):
            image = image
        elif isinstance(image, np.ndarray):
            image = Image.fromarray(cv2.cvtColor(image, cv2.COLOR_BGR2RGB))
        else:
            raise NotImplementedError(f'image with type {type(image)} not supported')

        image = self.procees_image(image)
        return {'image': image}

class AutoSizeHandler(DataHandler):
    def __init__(self, mode='resize', key_map_in=('image -> image', 'image_size -> size'), key_map_out=('image -> image', 'coord -> coord')):
        '''
        
        :param mode: ['full', 'resize', 'pad']
        '''
        super().__init__(key_map_in, key_map_out)
        self.mode = mode

    def handle(self, image, size):
        if self.mode == 'full':
            w, h = image.size
            coord = [h, w, 0, 0, h, w]
        elif self.mode == 'resize':
            image, coord = resize_crop_fix({'image': image}, size)
            image = image['image']
        elif self.mode == 'pad':
            image, coord = pad_crop_fix({'image': image}, size)
            image = image['image']
        else:
            raise NotImplementedError(f'mode {self.mode} not supported')
        coord = torch.tensor(coord, dtype=torch.float)
        return {'image': image, 'coord': coord}





from typing import Dict, List, Tuple, Union

import torch
from torch.serialization import FILE_LIKE
from io import BytesIO

from rainbowneko.utils import is_list
from .base import CkptFormat

import onnx
import onnxruntime


class ONNXFormat(CkptFormat):
    EXT = 'onnx'

    def __init__(self, inputs: Dict[str, Tuple] = {'input': (1, 3, 224, 224)}, output_names: List[str] = ['output'],
                 dynamic_axes: Union[Dict[str, Dict], str] = 'batch'):
        self.inputs, self.input_names, parsed_dynamic_axes = self.parse_inputs(inputs)
        self.output_names = output_names

        if len(parsed_dynamic_axes) > 0:
            self.dynamic_axes = parsed_dynamic_axes
            if dynamic_axes == 'batch':
                self.dynamic_axes.update({key: {0: 'batch'} for key in output_names})
            else:
                self.dynamic_axes.update(dynamic_axes)
        elif dynamic_axes == 'batch':
            self.dynamic_axes = {key: {0: 'batch'} for key in inputs.keys()}
            self.dynamic_axes.update({key: {0: 'batch'} for key in output_names})
        else:
            self.dynamic_axes = dynamic_axes

    def parse_inputs(self, inputs: Dict[str, Tuple]):
        '''
        {'input': (('batch', 1),3,224,224)} -> inputs={'input':(1,3,224,224)}, dynamic_axes={'input':{0:'batch'}}
        :param inputs:
        :return:
        '''
        input_tensors = []
        input_names = []
        dynamic_axes = {}

        for key, shape in inputs.items():
            dynamic = {}
            shape_clean = []
            for i, x in enumerate(shape):
                if is_list(x):
                    dynamic[i] = x[0]
                    shape_clean.append(x[1])
                else:
                    shape_clean.append(x)
            input_names.append(key)
            input_tensors.append(torch.randn(shape_clean))
            if len(dynamic) > 0:
                dynamic_axes[key] = dynamic
        return tuple(input_tensors), input_names, dynamic_axes

    def save_ckpt(self, sd_model: Dict[str, torch.nn.Module], save_f: FILE_LIKE, **kwargs):
        torch.onnx.export(sd_model['base'], self.inputs, save_f, input_names=self.input_names, output_names=self.output_names,
                          dynamic_axes=self.dynamic_axes, **kwargs)

    def load_ckpt(self, ckpt_f: FILE_LIKE, providers=['CPUExecutionProvider', 'CUDAExecutionProvider']):
        if isinstance(ckpt_f, BytesIO):
            ckpt_f = ckpt_f.getvalue()
        ort_session = onnxruntime.InferenceSession(ckpt_f, providers=providers)
        return ort_session



from typing import Dict, Any, Callable

import torch
from rainbowneko.parser.model import CfgPluginParser

from .base import BasicAction


class PrepareAction(BasicAction):
    def __init__(self, device='cuda', dtype=torch.float32, key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.device = torch.device(device)
        self.dtype = dtype

    def forward(self, in_preview=False, **states):
        if not in_preview:
            return {'device': self.device, 'dtype': self.dtype}


class BuildModelAction(BasicAction):
    def __init__(self, model_builder: Callable, key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.model_builder = model_builder

    def forward(self, device, in_preview=False, **states):
        if not in_preview:
            model = self.model_builder().to(device)
            model.eval()
            return {'model': model}


class BuildPluginAction(BasicAction):
    def __init__(self, parser: CfgPluginParser, key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.parser = parser

    def forward(self, model, device, in_preview=False, **states):
        if not in_preview:
            train_params, all_plugin_group = self.parser.get_params_group(model)
            model.eval()
            return {'all_plugin_group': all_plugin_group}


class ForwardAction(BasicAction):
    def to_dev(self, x, device, dtype):
        if isinstance(x, torch.Tensor):
            if torch.is_floating_point(x):
                return x.to(device, dtype=dtype)
            else:
                return x.to(device)
        else:
            return x

    def forward(self, input: Dict[str, Any], model, device, dtype, **states):
        with torch.no_grad(), torch.amp.autocast(device.type, dtype=dtype):
            input = {k:self.to_dev(v, device, dtype) for k,v in input.items()}
            output: Dict[str, Any] = model(**input)
        return {'output': output}


class VisPredAction(BasicAction):
    def forward(self, output: Dict[str, Any], **states):
        for k, v in output.items():
            if isinstance(v, torch.Tensor):
                print(f'{k} = {v.cpu()}')
            else:
                print(f'{k} = {v}')



from torch import nn
from typing import Dict, List, Tuple, Any

from rainbowneko.utils import KeyMapper, is_dict, is_list


class BaseWrapper(nn.Module):
    def post_init(self):
        pass

    @property
    def trainable_models(self) -> Dict[str, nn.Module]:
        raise NotImplementedError

    @property
    def trainable_parameters(self) -> List[nn.Parameter]:
        return [v for k, v in self.named_parameters() if v.requires_grad]

    @property
    def trainable_layers(self) -> List[nn.Module]:
        return self._trainable_layers

    @trainable_layers.setter
    def trainable_layers(self, layers: List[nn.Module]):
        self._trainable_layers = layers

    def update_model(self, step:int):
        pass

    def freeze_model(self):
        self.eval()

    def enable_gradient_checkpointing(self):
        pass

    def enable_xformers(self):
        pass

    def train(self, mode: bool = True):
        self.training = mode
        
        if hasattr(self, 'trainable_layers'):
            for layer in self.trainable_layers:
                if layer==self:
                    super().train(mode)
                else:
                    layer.train(mode)
        else:
            super().train(mode)
        return self

    def build_mapper(self, key_map, model=None, default=None):
        if is_dict(key_map) and len(key_map)>0:
            first_key = next(iter(key_map.values()))
            if is_dict(first_key) or isinstance(first_key, tuple) or is_list(first_key):
                return {k:KeyMapper(model, v or default) for k, v in key_map.items()}
        return KeyMapper(model, key_map or default)

    def get_map_data(self, key_mapper, data, ds_name:str=None) -> Tuple[Tuple, Dict[str, Any]]:
        if is_dict(key_mapper):
            model_args, model_kwargs = key_mapper[ds_name].map_data(data)
        else:
            model_args, model_kwargs = key_mapper.map_data(data)
        return model_args, model_kwargs

    def get_inputs_feed(self, key_mapper_in, model, kwargs, plugin_input={}, ds_name=None) -> Tuple[Tuple, Dict[str, Any]]:
        model_args, model_kwargs = self.get_map_data(key_mapper_in, kwargs, ds_name)

        input_all = dict(_args_=model_args, **model_kwargs, **plugin_input)

        if hasattr(model, 'input_feeder'):
            for feeder in model.input_feeder:
                feeder(input_all)
        return model_args, model_kwargs

class SingleWrapper(BaseWrapper):
    def __init__(self, model, key_map_in=None, key_map_out=None):
        super().__init__()
        self.model = model
        self.key_mapper_in = self.build_mapper(key_map_in, model, {0: 'image'})
        self.key_mapper_out = self.build_mapper(key_map_out, model, {'pred': 0})

    def forward(self, ds_name=None, plugin_input={}, **kwargs):
        model_args, model_kwargs = self.get_inputs_feed(self.key_mapper_in, self.model, kwargs, plugin_input, ds_name=ds_name)

        out = self.model(*model_args, **model_kwargs)
        if not isinstance(out, (tuple, dict)):
            out = (out,)
        return self.get_map_data(self.key_mapper_out, out, ds_name=ds_name)[1]

    @property
    def trainable_models(self) -> Dict[str, nn.Module]:
        return {'self':self}



from functools import partial

import torch
import torchvision
import torchvision.transforms as T
from timm.data import create_transform
from timm.data.constants import IMAGENET_DEFAULT_MEAN, IMAGENET_DEFAULT_STD
from torch import nn
from torch.nn import CrossEntropyLoss
from torchmetrics.classification import MulticlassAccuracy, MulticlassF1Score
from torchvision.transforms import InterpolationMode

from cfgs.py.train.classify import multi_class
from rainbowneko.ckpt_manager import ckpt_saver
from rainbowneko.evaluate import MetricGroup, MetricContainer
from rainbowneko.models.ema import ModelEMA
from rainbowneko.models.wrapper import DistillationWrapper
from rainbowneko.parser import CfgModelParser, neko_cfg
from rainbowneko.data import BaseDataset
from rainbowneko.data import FixedBucket
from rainbowneko.data.handler import HandlerChain, ImageHandler, LoadImageHandler, HandlerGroup
from rainbowneko.data.source import IndexSource, UnLabelSource
from rainbowneko.train.loss import LossContainer, PseudoLabelLoss
from rainbowneko.utils.lr_scheduler import fractional_warmup_schedule

num_classes = 10


def load_resnet(model, path=None):
    model.fc = nn.Linear(model.fc.in_features, num_classes)
    if path:
        model.load_state_dict(torch.load(path)['base'])
    return model

@neko_cfg
def make_cfg():
    return dict(
        _base_=[multi_class],

        model_part=CfgModelParser([
            dict(
                lr=1e-2,
                layers=['model_student'],
            )
        ]),

        ckpt_saver=dict(
            model=ckpt_saver(target_module='model_student')
        ),

        train=dict(
            train_epochs=100,
            save_step=2000,

            loss=dict(_replace_=True,
                dataset_S=LossContainer(CrossEntropyLoss(), key_map=('pred.pred_student -> 0', 'inputs.label -> 1')),
                dataset_U=LossContainer(PseudoLabelLoss(), key_map=('pred.pred_student -> pred', 'pred.pred_teacher -> pred_label')),
            ),

            metrics=dict(_replace_=True,
                dataset_S=MetricGroup(
                    acc=MetricContainer(MulticlassAccuracy(num_classes=num_classes), key_map=('pred.pred_student -> 0', 'inputs.label -> 1')),
                    f1=MetricContainer(MulticlassF1Score(num_classes=num_classes), key_map=('pred.pred_student -> 0', 'inputs.label -> 1')),
                ),
                dataset_U=None,
            ),
        ),

        model=dict(
            name='cifar-resnet18',
            wrapper=DistillationWrapper(_partial_=True, _replace_=True,
                key_map_in_student=dict(
                    dataset_S=('image -> 0',),
                    dataset_U=('image_strong -> 0',),
                ),
                key_map_in_teacher=dict(
                    dataset_S={},
                    dataset_U=('image_weak -> 0',),
                ),
                key_map_out=('pred_student -> pred_student', 'pred_teacher -> pred_teacher'),
                model_teacher=load_resnet(torchvision.models.resnet18()),
                model_student=load_resnet(torchvision.models.resnet18()),
                ema=ModelEMA(_partial_=True,
                    scheduler=fractional_warmup_schedule()
                )
            )
        ),

        data_train=dict(_replace_=True,
            dataset_S=partial(BaseDataset, batch_size=16, loss_weight=1.0,
                source=dict(
                    data_source1=IndexSource(
                        data=torchvision.datasets.cifar.CIFAR10(root=r'D:\others\dataset\cifar', train=True, download=True)
                    ),
                ),
                handler=HandlerChain(
                    load=LoadImageHandler(),
                    bucket=FixedBucket.handler, # bucket 会自带一些处理模块
                    image=ImageHandler(transform=create_transform(
                            input_size=224,
                            is_training=True,
                            color_jitter=0.4,
                            auto_augment='rand-m9-mstd0.5-inc1',
                            interpolation='bicubic',
                            re_prob=0.25,
                            re_mode='pixel',
                            re_count=1,
                            mean=IMAGENET_DEFAULT_MEAN,
                            std=IMAGENET_DEFAULT_STD,
                        )
                    )
                ),
                bucket=FixedBucket(target_size=224),
            ),
            dataset_U=partial(BaseDataset, batch_size=16, loss_weight=1.0,
                source=dict(
                    data_source1=UnLabelSource(
                        img_root=r'E:\dataset\k40\k40'
                    ),
                ),
                handler=HandlerGroup(
                    strong=HandlerChain(
                        load=LoadImageHandler(),
                        image=ImageHandler(transform=create_transform(
                                input_size=224,
                                is_training=True,
                                color_jitter=0.4,
                                auto_augment='rand-m9-mstd0.5-inc1',
                                interpolation='bicubic',
                                re_prob=0.25,
                                re_mode='pixel',
                                re_count=1,
                                mean=IMAGENET_DEFAULT_MEAN,
                                std=IMAGENET_DEFAULT_STD,
                            )
                        ), key_map_out=('image -> image_strong',)),
                    weak=HandlerChain(
                        load=LoadImageHandler(),
                        image=ImageHandler(transform=T.Compose([
                            T.Resize(256, interpolation=InterpolationMode.BICUBIC),
                            T.CenterCrop(224),
                            T.ToTensor(),
                            T.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
                        ])
                        ), key_map_out=('image -> image_weak',)),
                ),
                bucket=FixedBucket(target_size=224),
            )
        ),
    )



from functools import partial
from typing import Union, Callable, Any
from types import ModuleType

from hydra._internal.instantiate import _instantiate2

from .recursive_partial import RecursivePartial

####### patch hydra resolve target #######
target_stack = []

def getattr_call(obj: RecursivePartial, attr: str):
    if type(obj) is partial:
        obj = RecursivePartial(obj)
        obj.add_sub = True
        return obj(getattr, attr)
    elif type(obj) is RecursivePartial:
        obj.add_sub = True
        return obj(getattr, attr)
    else:
        return getattr(obj, attr)

def _resolve_target(
        target: Union[str, type, Callable[..., Any]], full_key: str
) -> Union[type, Callable[..., Any]]:
    """Resolve target string, type or callable into type or callable."""
    if isinstance(target, str):
        try:
            target = _instantiate2._locate(target)
        except Exception as e:
            msg = f"Error locating target '{target}', set env var HYDRA_FULL_ERROR=1 to see chained exception."
            if full_key:
                msg += f"\nfull_key: {full_key}"
            raise _instantiate2.InstantiationException(msg) from e
    elif _instantiate2._is_target(target):  # Recursive target resolve
        target = _instantiate2.instantiate_node(target)
    if not callable(target):
        msg = f"Expected a callable target, got '{target}' of type '{type(target).__name__}'"
        if full_key:
            msg += f"\nfull_key: {full_key}"
        raise _instantiate2.InstantiationException(msg)

    if type(target) is partial:
        target = RecursivePartial(target)
        target.add_sub = True
    elif target is getattr:
        target = getattr_call
    elif type(target) is RecursivePartial:
        target.add_sub = True

    return target

_instantiate2._resolve_target = _resolve_target
####### patch hydra resolve target #######

from .model import CfgModelParser, CustomModelParser, CfgWDModelParser, CfgPluginParser, CfgWDPluginParser
from .python_cfg import PythonCfgParser, get_rel_path, disable_neko_cfg
from .yaml_cfg import YamlCfgParser

def neko_cfg(func):
    parser = PythonCfgParser()
    return parser.compile_cfg(func)

def load_config(path: Union[str, ModuleType], remove_undefined=True):
    if isinstance(path, ModuleType):
        parser = PythonCfgParser()
    elif path.lower().endswith('.yaml'):
        parser = YamlCfgParser()
    elif path.lower().endswith('.py'):
        parser = PythonCfgParser()
    else:
        raise ValueError('Unsupported config file format: {}'.format(path))
    return parser, parser.load_config(path, remove_undefined)


def load_config_with_cli(path: Union[str, ModuleType], args_list=None, remove_undefined=True):
    if isinstance(path, ModuleType):
        parser = PythonCfgParser()
    elif path.lower().endswith('.yaml'):
        parser = YamlCfgParser()
    elif path.lower().endswith('.py'):
        parser = PythonCfgParser()
    else:
        raise ValueError('Unsupported config file format: {}'.format(path))
    return parser, parser.load_config_with_cli(path, args_list, remove_undefined)

def load_config_instant(path: str, remove_undefined=True):
    import hydra
    if path.lower().endswith('.yaml'):
        parser = YamlCfgParser()
    elif path.lower().endswith('.py'):
        parser = PythonCfgParser()
    else:
        raise ValueError('Unsupported config file format: {}'.format(path))
    cfg = parser.load_config(path, remove_undefined)
    return hydra.utils.instantiate(cfg)


from typing import Dict, Any

from .img_label import ImageLabelSource


class ImagePairSource(ImageLabelSource):
    def __init__(self, img_root, label_file, repeat=1, **kwargs):
        super().__init__(img_root=img_root, label_file=label_file, repeat=repeat)

    def __getitem__(self, index) -> Dict[str, Any]:
        img_id = self.img_ids[index]
        path = self.img_root / img_id
        label_img_name = self.label_dict.get(img_id, None)
        return {
            'id': img_id,
            'image': path,
            'label': self.img_root / label_img_name
        }



"""
train_ac.py
====================
    :Name:        train with accelerate
    :Author:      Dong Ziyi
    :Affiliation: HCP Lab, SYSU
    :Created:     10/03/2023
    :Licence:     Apache-2.0
"""

import argparse
import math
import os
import warnings
from functools import partial
from typing import Dict

import hydra
import torch
import torch.distributed as dist
import torch.utils.checkpoint
import torch.utils.data
from accelerate import Accelerator
from accelerate.utils import set_seed
from rainbowneko import _share
from rainbowneko.ckpt_manager import NekoSaver, NekoResumer
from rainbowneko.data import DataGroup, get_sampler, CacheableDataset
from rainbowneko.evaluate import EvaluatorGroup, MetricGroup
from rainbowneko.models.ema import ModelEMA
from rainbowneko.models.wrapper import BaseWrapper
from rainbowneko.parser import load_config_with_cli
from rainbowneko.train.loggers import LoggerGroup
from rainbowneko.utils import get_scheduler, mgcd, format_number, disable_hf_loggers, is_dict, xformers_available, maybe_DDP


class Trainer:
    weight_dtype_map = {"fp32": torch.float32, "fp16": torch.float16, "bf16": torch.bfloat16}

    def __init__(self, parser, cfgs_raw):
        torch.backends.cudnn.benchmark = True

        cfgs = hydra.utils.instantiate(cfgs_raw)
        self.cfgs = cfgs
        self.parser = parser

        self.init_context(cfgs_raw)
        self.build_loggers(cfgs_raw)

        self.build_ckpt_saver()
        self.build_model()

        for callback in _share.model_callbacks:
            callback(self.model_wrapper)

        # build dataset
        self.batch_size_list = []
        assert len(cfgs.data_train) > 0, "At least one dataset is need."
        loss_weights = {name:dataset.keywords["loss_weight"] for name, dataset in cfgs.data_train.items()}
        self.train_loader_group = DataGroup(
            {name:self.build_data(dataset, train=True) for name, dataset in cfgs.data_train.items()}, loss_weights
        )

        # calculate steps and epochs
        self.steps_per_epoch = len(self.train_loader_group.first_loader())
        if self.cfgs.train.train_epochs is not None:
            self.cfgs.train.train_steps = self.cfgs.train.train_epochs * self.steps_per_epoch // self.cfgs.train.gradient_accumulation_steps
        else:
            self.cfgs.train.train_epochs = math.ceil(self.cfgs.train.gradient_accumulation_steps * self.cfgs.train.train_steps / self.steps_per_epoch)

        self.build_optimizer_scheduler()
        self.model_wrapper.post_init()
        self.config_model()
        self.build_loss()

        with torch.no_grad():
            self.build_ema()

        self.load_resume(self.cfgs.train.resume)

        torch.backends.cuda.matmul.allow_tf32 = cfgs.allow_tf32

        if self.cfgs.train.metrics is not None:
            self.metric_train = self.cfgs.train.metrics
            if is_dict(self.metric_train):
                for metric in self.metric_train.values():
                    if metric is not None:
                        metric.to(self.accelerator.device)
            else:
                self.metric_train.to(self.accelerator.device)
        else:
            self.metric_train = None

        if self.cfgs.evaluator is not None:
            self.evaluator = self.build_evaluator(self.cfgs.evaluator)
            self.evaluator.to(self.accelerator.device)
        else:
            self.evaluator = None

        self.prepare()

    @property
    def device(self):
        return self.accelerator.device

    @property
    def is_local_main_process(self):
        return self.accelerator.is_local_main_process

    def init_context(self, cfgs_raw):
        self.accelerator = Accelerator(
            gradient_accumulation_steps=self.cfgs.train.gradient_accumulation_steps,
            mixed_precision=self.cfgs.mixed_precision,
            step_scheduler_with_optimizer=False,
        )

        self.local_rank = int(os.environ.get("LOCAL_RANK", -1))
        self.world_size = self.accelerator.num_processes
        _share.local_rank = self.local_rank
        _share.world_size = self.world_size
        _share.device = self.device

        set_seed(self.cfgs.seed + self.local_rank)

    def build_loggers(self, cfgs_raw):
        if self.is_local_main_process:
            self.exp_dir = self.cfgs.exp_dir
            os.makedirs(os.path.join(self.exp_dir, "ckpts/"), exist_ok=True)
            self.parser.save_configs(cfgs_raw, self.exp_dir)
            self.loggers: LoggerGroup = LoggerGroup([builder(exp_dir=self.exp_dir) for builder in self.cfgs.logger])
        else:
            self.loggers: LoggerGroup = LoggerGroup([builder(exp_dir=None) for builder in self.cfgs.logger])

        _share.loggers = self.loggers
        self.min_log_step = mgcd(*([item.log_step for item in self.loggers.logger_list]))

        self.loggers.info(f"world size (num GPUs): {self.world_size}")
        self.loggers.info(f"accumulation: {self.cfgs.train.gradient_accumulation_steps}")

        disable_hf_loggers(self.is_local_main_process)

    def prepare(self):
        # Prepare everything with accelerator.
        prepare_name_list, prepare_obj_list = [], []
        for k, v in self.model_wrapper.trainable_models.items():
            prepare_obj_list.append(v)
            prepare_name_list.append(k)
        N_trainable_models = len(prepare_obj_list)

        if hasattr(self, "optimizer"):
            prepare_obj_list.extend([self.optimizer, self.lr_scheduler] if self.lr_scheduler else [self.optimizer])
            prepare_name_list.extend(["optimizer", "lr_scheduler"] if self.lr_scheduler else ["optimizer"])

        prepare_obj_list.extend(self.train_loader_group.loader_dict.values())
        prepared_obj = self.accelerator.prepare(*prepare_obj_list)

        # prepared model
        if prepare_name_list[0] == "self":  # entire model is trainable
            self.model_wrapper = prepared_obj[0]
        else:
            for name, obj in zip(prepare_name_list[:N_trainable_models], prepared_obj[:N_trainable_models]):
                setattr(self.model_wrapper, name, obj)
        prepare_name_list = prepare_name_list[N_trainable_models:]
        prepared_obj = prepared_obj[N_trainable_models:]

        # prepared dataset
        ds_num = len(self.train_loader_group.loader_dict)
        for i, (k,v) in enumerate(self.train_loader_group.loader_dict.items()):
            self.train_loader_group.loader_dict[k] = prepared_obj[-ds_num+i]
        prepared_obj = prepared_obj[:-ds_num]

        # prepared optimizer and scheduler
        for name, obj in zip(prepare_name_list, prepared_obj):
            setattr(self, name, obj)

        self.model_wrapper: BaseWrapper = self.model_wrapper.to(self.device)
        if self.cfgs.model.force_cast_precision:
            self.model_wrapper.to(dtype=self.weight_dtype)

    def scale_lr(self, parameters):
        bs = sum(self.batch_size_list)
        scale_factor = bs * self.world_size * self.cfgs.train.gradient_accumulation_steps
        for param in parameters:
            if "lr" in param:
                param["lr"] *= scale_factor

    def build_model(self):
        self.model_wrapper = self.cfgs.model.wrapper()

        self.model_wrapper.requires_grad_(False)
        self.model_wrapper.eval()
        self.weight_dtype = self.weight_dtype_map.get(self.cfgs.mixed_precision, torch.float32)

    def build_ema(self):
        if self.cfgs.model.ema is not None:
            self.ema_model: ModelEMA = self.cfgs.model.ema(self.model_wrapper)

    def build_loss(self):
        criterion = self.cfgs.train.loss
        if is_dict(criterion):
            self.criterion = {name:loss.to(self.device) for name, loss in criterion.items()}
        else:
            self.criterion = criterion.to(self.device)

    def build_ckpt_saver(self):
        if self.is_local_main_process:
            self.ckpt_saver: Dict[str, NekoSaver] = self.cfgs.ckpt_saver
            self.ckpt_dir = os.path.join(self.exp_dir, "ckpts")
            for ckpt_saver in self.ckpt_saver.values():
                ckpt_saver.prefix = self.ckpt_dir

    def build_evaluator(self, cfgs_eval):
        def build_one(cfgs_eval_one):
            dataset_cfg = cfgs_eval_one.keywords.pop('dataset', None)
            if dataset_cfg is None:
                evaluator = cfgs_eval_one(trainer=self)
            else:
                val_loader = self.build_data(dataset_cfg, train=False)
                evaluator = cfgs_eval_one(trainer=self, data_loader=val_loader)
            return evaluator

        if cfgs_eval is not None:
            if is_dict(cfgs_eval):
                evaluator_dict = {name:build_one(cfg) for name, cfg in cfgs_eval.items()}
                return EvaluatorGroup(loggers=self.loggers, evaluator_dict=evaluator_dict)
            else:
                return build_one(cfgs_eval)
        else:
            return None

    @property
    def model_raw(self):
        return maybe_DDP(self.model_wrapper)

    def config_model(self):
        if self.cfgs.model.enable_xformers:
            if xformers_available:
                self.model_wrapper.enable_xformers()
            else:
                warnings.warn("xformers is not available. Make sure it is installed correctly")

        if self.cfgs.model.gradient_checkpointing:
            self.model_wrapper.enable_gradient_checkpointing()

    @torch.no_grad()
    def load_resume(self, resumer: NekoResumer):
        if resumer is not None:
            resumer.load_to(self.model_wrapper, getattr(self, "ema_model", None))

    def to_dev(self, x):
        if isinstance(x, torch.Tensor):
            if torch.is_floating_point(x):
                return x.to(self.device, dtype=self.weight_dtype)
            else:
                return x.to(self.device)
        else:
            return x

    def boardcast_main(self, data):
        obj = [data]
        dist.broadcast_object_list(obj, src=0)
        return obj[0]

    def all_gather(self, data):
        if not hasattr(self, 'gloo_group'): # Transfer data on cpu
            self.gloo_group = dist.new_group(backend='gloo')
        gathered_objects = [None for _ in range(self.world_size)]
        dist.all_gather_object(gathered_objects, data, group=self.gloo_group)
        return gathered_objects

    def build_dataset(self, data_builder: partial):
        batch_size = data_builder.keywords.pop("batch_size")
        self.batch_size_list.append(batch_size)

        dataset = data_builder()
        dataset.build_bucket(bs=batch_size, world_size=self.world_size)
        if isinstance(dataset, CacheableDataset):
            dataset.build_cache(self.model_wrapper, self.all_gather)
        self.loggers.info(f"len(dataset): {len(dataset)}")

        return dataset, batch_size

    def build_data(self, data_builder: partial, train=True) -> torch.utils.data.DataLoader:
        drop_last = data_builder.keywords.pop("drop_last", True)
        dataset, batch_size = self.build_dataset(data_builder)

        # Pytorch Data loader
        sampler = get_sampler(train)(
            dataset,
            num_replicas=self.world_size,
            rank=self.local_rank,
            shuffle=train and dataset.bucket.can_shuffle,
        )
        loader = torch.utils.data.DataLoader(
            dataset,
            batch_size=batch_size,
            num_workers=self.cfgs.train.workers,
            sampler=sampler,
            collate_fn=dataset.collate_fn,
            drop_last=drop_last,
        )
        return loader

    def get_param_group_train(self):
        # make model part and plugin
        if self.cfgs.model_part is None:
            train_params, train_layers = [], []
        else:
            train_params, train_layers = self.cfgs.model_part.get_params_group(self.model_wrapper)
        self.model_wrapper.trainable_layers = train_layers

        if self.cfgs.model_plugin is None:
            train_params_plugin, self.all_plugin = [], {}
        else:
            train_params_plugin, self.all_plugin = self.cfgs.model_plugin.get_params_group(self.model_wrapper)
        train_params += train_params_plugin

        N_params_unet = format_number(sum(sum(x.numel() for x in p["params"]) for p in train_params))
        self.loggers.info(f"model trainable params: {N_params_unet}")

        return train_params

    def build_optimizer_scheduler(self):
        # set optimizer
        parameters = self.get_param_group_train()

        if len(parameters) > 0:  # do fine-tuning
            cfg_opt = self.cfgs.train.optimizer
            if self.cfgs.train.scale_lr:
                self.scale_lr(parameters)
            assert isinstance(cfg_opt, partial), f"optimizer config should be partial."
            self.optimizer = cfg_opt(params=parameters)
            self.lr_scheduler = get_scheduler(self.cfgs.train.scheduler, self.optimizer, self.cfgs.train.train_steps)

    def train(self, loss_ema=0.93):
        acc_steps = self.cfgs.train.gradient_accumulation_steps
        total_batch_size = sum(self.batch_size_list) * self.world_size * acc_steps

        self.loggers.info("***** Running training *****")
        self.loggers.info(
            f"  Num batches each epoch = {[len(loader) for loader in self.train_loader_group.loader_dict.values()]}"
        )
        self.loggers.info(f"  Num Steps = {self.cfgs.train.train_steps}")
        self.loggers.info(f"  Instantaneous batch size per device = {sum(self.batch_size_list)}")
        self.loggers.info(f"  Total train batch size (w. parallel, distributed & accumulation) = {total_batch_size}")
        self.loggers.info(f"  Gradient Accumulation steps = {acc_steps}")
        self.global_step = 0
        self.real_step = 0
        if self.cfgs.train.resume is not None:
            self.global_step = self.cfgs.train.resume.start_step * acc_steps
            self.lr_scheduler.step(self.cfgs.train.resume.start_step)

        self.model_wrapper.train()
        loss_sum = None
        for data_dict in self.train_loader_group:
            loss, pred_dict, inputs_dict = self.train_one_step(data_dict)
            loss_sum = loss if loss_sum is None else (loss_ema * loss_sum + (1 - loss_ema) * loss)

            self.global_step += 1
            acc_step = self.global_step%acc_steps
            self.real_step = max(1, self.global_step//acc_steps)
            if self.is_local_main_process:
                if self.real_step % self.cfgs.train.save_step == 0 and acc_step==acc_steps-1:
                    self.save_model()
                if self.global_step % self.min_log_step == 0:
                    # get learning rate from optimizer
                    lr_model = self.optimizer.param_groups[0]["lr"] if hasattr(self, "optimizer") else 0.0
                    if acc_steps>1:
                        log_step={
                            "format": "[{}/{}]({}/{})",
                            "data": [self.real_step, self.cfgs.train.train_steps, self.global_step%acc_steps, acc_steps],
                        }
                    else:
                        log_step={
                            "format": "[{}/{}]",
                            "data": [self.real_step, self.cfgs.train.train_steps],
                        }
                    log_data = {
                        "train/Step": log_step,
                        "train/Epoch": {
                            "format": "[{}/{}]<{}/{}>",
                            "data": [
                                self.global_step // self.steps_per_epoch, self.cfgs.train.train_epochs,
                                self.global_step % self.steps_per_epoch, self.steps_per_epoch,
                            ],
                        },
                        "train/LR_model": {"format": "{:.2e}", "data": [lr_model]},
                        "train/Loss": {"format": "{:.5f}", "data": [loss_sum]},
                    }
                    if self.metric_train is not None:
                        if is_dict(self.metric_train):
                            for ds_name in pred_dict.keys():
                                if self.metric_train[ds_name] is not None:
                                    self.metric_train[ds_name].reset()
                                    self.metric_train[ds_name].update(pred_dict[ds_name], inputs_dict[ds_name])
                                    metrics_dict = self.metric_train[ds_name].finish(lambda x:x, self.is_local_main_process)
                                    log_data.update(MetricGroup.format(metrics_dict, prefix=f'{ds_name}/'))
                        else:
                            self.metric_train.reset()
                            for ds_name in pred_dict.keys():
                                self.metric_train.update(pred_dict[ds_name], inputs_dict[ds_name])
                            metrics_dict = self.metric_train.finish(lambda x:x, self.is_local_main_process)
                            log_data.update(MetricGroup.format(metrics_dict))
                    self.loggers.log(
                        datas=log_data,
                        step=self.real_step,
                    )

            if self.evaluator is not None and acc_step==acc_steps-1:
                self.evaluator.evaluate(self.real_step, self.model_raw)

            if self.real_step >= self.cfgs.train.train_steps and acc_step==acc_steps-1:
                break

        self.wait_for_everyone()
        if self.is_local_main_process:
            self.save_model()

    def train_one_step(self, data_dict):
        v_proc = lambda v: v.detach() if isinstance(v, torch.Tensor) else v

        pred_dict, inputs_dict = {}, {}
        loss_all = []
        with self.accelerator.accumulate(self.model_wrapper):
            for ds_name, data in data_dict.items():
                input_datas = {k: self.to_dev(v) for k, v in data.items() if k != "plugin_input"}
                if "plugin_input" in data:
                    input_datas["plugin_input"] = {k: self.to_dev(v) for k, v in data["plugin_input"].items()}

                model_pred = self.model_wrapper(ds_name, **input_datas)
                pred_dict[ds_name] = {k: v_proc(v) for k, v in model_pred.items()}
                inputs_dict[ds_name] = {k: v_proc(v) for k, v in input_datas.items()}
                loss = self.get_loss(ds_name, model_pred, input_datas)
                loss_all.append(loss.item())
                self.accelerator.backward(loss, retain_graph=self.cfgs.train.retain_graph)

            if hasattr(self, "optimizer"):
                if self.cfgs.train.max_grad_norm and self.accelerator.sync_gradients:  # fine-tuning
                    if hasattr(self.model_wrapper, "trainable_parameters"):
                        clip_param = self.model_wrapper.trainable_parameters
                    else:
                        clip_param = self.model_wrapper.module.trainable_parameters
                    self.accelerator.clip_grad_norm_(clip_param, self.cfgs.train.max_grad_norm)
                self.optimizer.step()
                if self.lr_scheduler:
                    self.lr_scheduler.step()
                self.optimizer.zero_grad(set_to_none=self.cfgs.train.set_grads_to_none)

            self.model_raw.update_model(self.real_step)  # Some model may update by step
            self.update_ema()
        return sum(loss_all), pred_dict, inputs_dict

    def get_loss(self, ds_name, model_pred, inputs):
        weight = self.train_loader_group.get_loss_weights(ds_name)
        if is_dict(self.criterion):
            loss = self.criterion[ds_name](model_pred, inputs).mean()
        else:
            loss = self.criterion(model_pred, inputs).mean()
        return loss * weight

    def update_ema(self):
        if hasattr(self, "ema_model"):
            self.ema_model.step(self.model_raw.named_parameters())

    def save_model(self, from_raw=False):
        NekoSaver.save_all(
            self.model_raw,
            plugin_groups=self.all_plugin,
            cfg=self.ckpt_saver,
            model_ema=getattr(self, "ema_model", None),
            name_template=f'{{}}-{self.real_step}',
        )

        self.loggers.info(f"Saved state, step: {self.real_step}")

    def wait_for_everyone(self):
        self.accelerator.wait_for_everyone()

def neko_train():
    import subprocess
    parser = argparse.ArgumentParser(description='RainbowNeko Launcher')
    parser.add_argument('--launch_cfg', type=str, default='cfgs/launcher/multi.yaml')
    args, train_args = parser.parse_known_args()

    subprocess.run(["accelerate", "launch", '--config_file', args.launch_cfg, "-m",
                    "rainbowneko.train.trainer.trainer_ac"] + train_args, check=True)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="RainbowNeko Trainer")
    parser.add_argument("--cfg", type=str, default=None, required=True)
    args, cfg_args = parser.parse_known_args()

    parser, conf = load_config_with_cli(args.cfg, args_list=cfg_args)  # skip --cfg
    trainer = Trainer(parser, conf)
    trainer.train()



"""
lora_layers.py
====================
    :Name:        lora layers
    :Author:      Dong Ziyi
    :Affiliation: HCP Lab, SYSU
    :Created:     09/04/2023
    :Licence:     Apache-2.0
"""

import math

import torch
from einops import einsum
from torch import nn
from torch.nn import functional as F
from torch.nn.modules.utils import _pair

from .lora_base_patch import LoraBlock


class LoraLayer(LoraBlock):
    def __init__(self, lora_id: int, host, rank=1, dropout=0.1, alpha=1.0, bias=False, alpha_auto_scale=True, **kwargs):
        super().__init__(lora_id, host, rank, dropout, alpha=alpha, bias=bias, alpha_auto_scale=alpha_auto_scale, **kwargs)

    class LinearLayer(LoraBlock.LinearLayer):
        def __init__(self, host: nn.Linear, rank, bias, block):
            super().__init__(host, rank, bias, block)
            if isinstance(self.rank, float):
                self.rank = max(round(host.out_features * self.rank), 1)

            self.in_features = host.in_features
            self.out_features = host.out_features
            self.W_down = nn.Parameter(torch.empty(self.rank, host.in_features))
            self.W_up = nn.Parameter(torch.empty(host.out_features, self.rank))
            if bias:
                self.bias = nn.Parameter(torch.empty(host.out_features))
            else:
                self.register_parameter('bias', None)

        def extra_repr(self) -> str:
            return f'in_features={self.in_features}, rank={self.rank}, out_features={self.out_features}'

        def reset_parameters(self):
            nn.init.kaiming_uniform_(self.W_down, a=math.sqrt(5))
            nn.init.zeros_(self.W_up)
            if self.bias:
                nn.init.zeros_(self.bias)

        def get_weight(self):
            return torch.mm(self.W_up, self.W_down)

        def get_bias(self):
            return self.bias

        def forward(self, x, weight, bias=None):
            # make it faster
            x_shape = x.shape
            if bias is None:
                return torch.mm(x.view(-1, x_shape[-1]), weight.transpose(0, 1)).view(*x_shape[:-1], -1)
            else:
                return torch.mm(x.view(-1, x_shape[-1]), weight.transpose(0, 1)).view(*x_shape[:-1], -1) + bias
            # return F.linear(x, weight, bias) # linear is slow

        def get_collapsed_param(self):
            w = self.W_up.data @ self.W_down.data
            b = self.bias.data if self.bias else None
            return w, b

    class Conv2dLayer(LoraBlock.Conv2dLayer):
        def __init__(self, host: nn.Conv2d, rank, bias, block):
            super().__init__(host, rank, bias, block)
            if isinstance(self.rank, float):
                self.rank = max(round(host.out_channels * self.rank), 1)

            self.in_channels = host.in_channels
            self.out_channels = host.out_channels
            self.kernel_size = host.kernel_size

            self.W_down = nn.Parameter(torch.empty(self.rank, host.in_channels, *host.kernel_size))
            self.W_up = nn.Parameter(torch.empty(host.out_channels, self.rank, 1, 1))
            if bias:
                self.bias = nn.Parameter(torch.empty(host.out_channels))
            else:
                self.register_parameter('bias', None)

            self.stride = host.stride
            self.padding = host.padding
            self.dilation = host.dilation
            self.groups = host.groups
            self.padding_mode = host.padding_mode
            self._reversed_padding_repeated_twice = host._reversed_padding_repeated_twice

        def extra_repr(self):
            s = ('{in_channels}, {out_channels}, rank={rank}, kernel_size={kernel_size}'
                 ', stride={stride}')
            if self.padding != (0,) * len(self.padding):
                s += ', padding={padding}'
            if self.dilation != (1,) * len(self.dilation):
                s += ', dilation={dilation}'
            if self.groups != 1:
                s += ', groups={groups}'
            if self.bias is None:
                s += ', bias=False'
            if self.padding_mode != 'zeros':
                s += ', padding_mode={padding_mode}'
            return s.format(**self.__dict__)

        def reset_parameters(self):
            nn.init.kaiming_uniform_(self.W_down, a=math.sqrt(5))
            nn.init.zeros_(self.W_up)
            if self.bias:
                nn.init.zeros_(self.bias)

        def get_weight(self):
            return einsum(self.W_up, self.W_down, 'o r ..., r i ... -> o i ...')

        def get_bias(self):
            return self.bias if self.bias else None

        def forward(self, x, weight, bias=None):
            if self.padding_mode != 'zeros':
                return F.conv2d(F.pad(input, self._reversed_padding_repeated_twice, mode=self.padding_mode),
                                weight, bias, self.stride, _pair(0), self.dilation, self.groups)
            return F.conv2d(x, weight, bias, self.stride, self.padding, self.dilation, self.groups)

        def get_collapsed_param(self):
            w = einsum(self.W_up.data, self.W_down.data, 'o r ..., r i ... -> o i ...')
            b = self.bias.data if self.bias else None
            return w, b

if __name__ == '__main__':
    from timm.models.swin_transformer_v2 import swinv2_base_window8_256

    model = swinv2_base_window8_256()
    lora_layers = LoraLayer.wrap_model(0, model)
    print(model)


from ..format import CkptFormat


class CkptSource:

    def get(self, name: str, format: CkptFormat, prefix=None, **kwargs):
        raise NotImplementedError

    def put(self, name: str, data, format: CkptFormat, prefix=None):
        raise NotImplementedError

    def __repr__(self):
        return f'{self.__class__.__name__}()'



import torch
from torch import nn
from torch.nn.parallel import DistributedDataParallel as DDP

try:
    import xformers

    xformers_available = True
except:
    xformers_available = False

def remove_all_hooks(model: nn.Module) -> None:
    for name, child in model.named_modules():
        child._forward_hooks.clear()
        child._forward_pre_hooks.clear()
        child._backward_hooks.clear()

def remove_layers(model: nn.Module, layer_class):
    named_modules = {k:v for k, v in model.named_modules()}
    for k, v in named_modules.items():
        if isinstance(v, layer_class):
            parent, name = named_modules[k.rsplit('.', 1)]
            delattr(parent, name)
            del v

def hook_compile(model):
    named_modules = {k:v for k, v in model.named_modules()}

    for name, block in named_modules.items():
        if len(block._forward_hooks)>0:
            for hook in block._forward_hooks.values():  # 从前往后执行
                old_forward = block.forward

                def new_forward(*args, **kwargs):
                    result = old_forward(*args, **kwargs)
                    hook_result = hook(block, args, result)
                    if hook_result is not None:
                        result = hook_result
                    return result

                block.forward = new_forward

        if len(block._forward_pre_hooks)>0:
            for hook in list(block._forward_pre_hooks.values())[::-1]:  # 从前往后执行
                old_forward = block.forward

                def new_forward(*args, **kwargs):
                    result = hook(block, args)
                    if result is not None:
                        if not isinstance(result, tuple):
                            result = (result,)
                    else:
                        result = args
                    return old_forward(*result, **kwargs)

                block.forward = new_forward
    remove_all_hooks(model)

def _convert_cpu(t):
    return t.to('cpu') if t.device.type == 'cuda' else t

def _convert_cuda(t):
    return t.to('cuda') if t.device.type == 'cpu' else t

def to_cpu(model):
    model._apply(_convert_cpu)
    torch.cuda.empty_cache()
    torch.cuda.synchronize()

def to_cuda(model):
    model._apply(_convert_cuda)

def split_module_name(layer_name):
    name_split = layer_name.rsplit('.', 1)
    if len(name_split) == 1:
        parent_name, host_name = '', name_split[0]
    else:
        parent_name, host_name = name_split
    return parent_name, host_name

def maybe_DDP(model):
    if isinstance(model, DDP):
        return model.module
    else:
        return model

def zero_module(module):
    """
    Zero out the parameters of a module and return it.
    """
    for p in module.parameters():
        p.detach().zero_()
    return module


import time
import warnings
from omegaconf import OmegaConf
import importlib

def times(a,b):
    warnings.warn(f"${{times:{a},{b}}} is deprecated and will be removed in the future. Please use ${{hcp.eval:{a}*{b}}} instead.", DeprecationWarning)
    return a*b

OmegaConf.register_new_resolver("times", times)

OmegaConf.register_new_resolver("neko.eval", lambda exp: eval(exp))
OmegaConf.register_new_resolver("neko.time", lambda format="%Y-%m-%d-%H-%M-%S": time.strftime(format))

def get(name):
    module = name.split('.', 1)[0]
    exec(f'import {module}')
    return eval(name)

OmegaConf.register_new_resolver("neko.get", get) # get python object

#OmegaConf.register_new_resolver("neko.dtype", lambda dtype: dtype_dict.get(dtype, torch.float32))


from typing import Dict

from .base import NekoLoader


class NekoResumer:
    def __init__(self, loader: Dict[str, NekoLoader], start_step: int, loader_ema: Dict[str, NekoLoader] = None):
        self.loader = loader
        self.start_step = start_step
        self.loader_ema = loader_ema

    def load_to(self, model, ema_model=None):
        NekoLoader.load_all(model, self.loader)
        if ema_model is not None:
            NekoLoader.load_all(ema_model, self.loader_ema)



from copy import deepcopy
from typing import Dict, Iterable, Union, Tuple

import torch
from rainbowneko.utils.lr_scheduler import SchedulerType
from torch import nn

ModelParamsType = Union[Dict[str, nn.Parameter], Iterable[Tuple[str, nn.Parameter]], nn.Module]


class ModelEMA:
    def __init__(self, model: nn.Module, scheduler: SchedulerType, start_step=0, interval=1):
        with torch.no_grad():
            self.model = deepcopy(model)
            self.train_params = {name: p for name, p in model.named_parameters()}

        self.scheduler = scheduler
        self.optimization_step = start_step
        self.interval = interval

        self.foreach = hasattr(torch, '_foreach_copy_')

    @torch.no_grad()
    def step(self, model: nn.Module, step: int = 0):
        self.optimization_step += 1

        if step % self.interval == 0:
            # Compute the decay factor for the exponential moving average.
            one_minus_decay = 1 - self.scheduler(step)

            if self.foreach:
                params_grad = [
                    param.to(s_param.device) for s_param, param in zip(self.model.parameters(), model.parameters())
                    if param.requires_grad
                ]
                s_params_grad = [
                    s_param for s_param, param in zip(self.model.parameters(), model.parameters()) if param.requires_grad
                ]

                torch._foreach_sub_(s_params_grad, torch._foreach_sub(s_params_grad, params_grad), alpha=one_minus_decay)

            else:
                for s_param, param in zip(self.model.parameters(), model.parameters()):
                    s_param.sub_(one_minus_decay * (s_param - param.to(s_param.device)))

        # torch.cuda.empty_cache()

    def __call__(self, *args, **kwargs):
        return self.model(*args, **kwargs)

    def copy_to(self, model: nn.Module) -> None:
        if self.foreach:
            parameters = list(model.parameters())
            torch._foreach_copy_(
                [param.data for param in parameters],
                [s_param.to(param.device).data for s_param, param in zip(self.model.parameters(), parameters)],
            )
        else:
            for s_param, param in zip(self.model.parameters(), model.parameters()):
                param.data.copy_(s_param.to(param.device).data)

    def to(self, device=None, dtype=None):
        with torch.no_grad():
            self.model.to(device, dtype)
            self.train_params = {name: p for name, p in self.model.named_parameters()}
        return self

    def state_dict(self) -> Dict[str, torch.Tensor]:
        return self.model.state_dict()

    def load_state_dict(self, state: Dict[str, torch.Tensor], prefix=None):
        for k, v in state:
            if prefix is None:
                if k in self.train_params:
                    self.train_params[k].data = v
            else:
                if k in self.train_params and k.startswith(prefix):
                    self.train_params[k].data = v


class ParameterEMA:
    def __init__(self, params: ModelParamsType, scheduler: SchedulerType, start_step=0, interval=1):
        with torch.no_grad():
            self.params = self.get_params(params)

        self.scheduler = scheduler
        self.optimization_step = start_step
        self.interval = interval

        self.foreach = hasattr(torch, '_foreach_copy_')

    def get_params(self, model_or_params: ModelParamsType, grad_only=True):
        if isinstance(model_or_params, nn.Module):
            return {name: p for name, p in model_or_params.named_parameters() if not grad_only or p.requires_grad}
        elif isinstance(model_or_params, dict):
            return {name: p for name, p in model_or_params.items() if not grad_only or p.requires_grad}
        else:
            return {name: p for name, p in model_or_params if not grad_only or p.requires_grad}

    @torch.no_grad()
    def step(self, params: ModelParamsType, step: int = 0):
        self.optimization_step += 1
        params = self.get_params(params)

        if step % self.interval == 0:
            # Compute the decay factor for the exponential moving average.
            one_minus_decay = 1 - self.scheduler(step)

            if self.foreach:
                params_grad = [param.to(self.params[name].device) for name, param in params.items()]
                s_params_grad = [self.params[name] for name in params.keys()]

                torch._foreach_sub_(s_params_grad, torch._foreach_sub(s_params_grad, params_grad), alpha=one_minus_decay)

            else:
                for name, param in params.items():
                    s_param = self.params[name]
                    s_param.sub_(one_minus_decay * (s_param - param.to(s_param.device)))

    def copy_to(self, params: ModelParamsType) -> None:
        params = self.get_params(params)
        if self.foreach:
            torch._foreach_copy_(
                [param.data for param in params.values()],
                [self.params[name].to(param.device).data for name, param in params.items()],
            )
        else:
            for name, param in params.items():
                s_param = self.params[name]
                param.data.copy_(s_param.to(param.device).data)

    def to(self, device=None, dtype=None):
        with torch.no_grad():
            self.params = {name: p.to(device, dtype) for name, p in self.params.items()}
        return self

    def state_dict(self) -> Dict[str, torch.Tensor]:
        return {name: p.data for name, p in self.params.items()}

    def load_state_dict(self, state: Dict[str, torch.Tensor], prefix=None):
        for k, v in state:
            if prefix is None:
                if k in self.params:
                    self.params[k].data = v
            else:
                if k in self.params and k.startswith(prefix):
                    self.params[k].data = v



import bisect
from pathlib import Path
from typing import Dict, List, Tuple, Any

from rainbowneko.utils import Path_Like
from rainbowneko.utils.img_size_tool import get_image_size
from contextlib import contextmanager


class DataSource:
    def __init__(self, repeat=1, **kwargs):
        self.repeat = repeat

    def __getitem__(self, index) -> Dict[str, Any]:
        return index

    def __len__(self):
        raise NotImplementedError()


class ComposeDataSource(DataSource):
    def __init__(self, source_list: List[DataSource]):
        self.source_list = source_list

        offsets = [0]
        for source in self.source_list:
            offsets.append(offsets[-1] + len(source))
        self._offsets = offsets

        self._return_source = False

    @contextmanager
    def return_source(self):
        self._return_source = True
        yield
        self._return_source = False

    def get_source_by_index(self, index) -> DataSource:
        if index < 0 or index >= len(self):
            raise IndexError('Index out of range')

        # 使用二分查找来找到正确的序列
        seq_index = bisect.bisect_right(self._offsets, index) - 1
        return self.source_list[seq_index]

    def __getitem__(self, index) -> Dict[str, Any]:
        if index < 0 or index >= len(self):
            raise IndexError('Index out of range')

        # 使用二分查找来找到正确的序列
        seq_index = bisect.bisect_right(self._offsets, index) - 1
        index_within_seq = index - self._offsets[seq_index]
        source = self.source_list[seq_index]
        if self._return_source:
            return source[index_within_seq], source
        else:
            return source[index_within_seq]

    def __len__(self):
        return self._offsets[-1]


class VisionDataSource(DataSource):
    def __init__(self, img_root: Path_Like, repeat=1, **kwargs):
        super().__init__(repeat=repeat)
        self.img_root = Path(img_root)

    def get_image_size(self, data: Dict[str, Any]) -> Tuple[int, int]:
        return get_image_size(data['image'])



from .base import DataHandler

class LabelHandler(DataHandler):
    def __init__(self, transform=None, key_map_in=('label -> label',), key_map_out=('label -> label',)):
        super().__init__(key_map_in, key_map_out)
        self.transform = transform

    def handle(self, label):
        label = self.transform(label)
        return {'label': label}



# for training loggers
loggers = None

# for multi-gpu training
local_rank = -1
world_size = 1
device = 'cuda'

# for model related methods
model_callbacks = []

def register_model_callback(callback):
    model_callbacks.append(callback)
    return callback


from .base import BasicAction, LoopAction, FromMemory, Actions, LambdaAction
from .classify import VisClassAction
from .data import HandlerAction, DataLoaderAction
from .evaluate import MetricAction
from .io import LoadImageAction, LoadModelAction, FeedAction
from .model import ForwardAction, VisPredAction, BuildModelAction, PrepareAction, BuildPluginAction



import ast
import importlib
import inspect
import linecache
import os
import random
import re
import shutil
import string
from contextlib import nullcontext
from types import ModuleType
from typing import Union

from omegaconf import OmegaConf

from .yaml_cfg import YamlCfgParser


def get_rel_path(path):
    if not isinstance(path, str):
        path = path.__file__
    current_dir = os.getcwd()
    return os.path.relpath(path, current_dir)


def random_filename(length=8, suffix=".py"):
    chars = string.ascii_letters + string.digits  # 字母 + 数字
    random_part = ''.join(random.choices(chars, k=length))
    return f"{random_part}{suffix}"


disable_neko_cfg = nullcontext()


class CallTransformer(ast.NodeTransformer):
    transform_parent = (ast.Call, ast.Expr, ast.Dict, ast.List, ast.Return, ast.Assign)
    node_skip = ast.Lambda

    def __init__(self):
        self.parent_stack = []

    def visit(self, node):
        if isinstance(node, self.node_skip):
            return node

        # 在遍历到每个节点时，将当前节点推入栈
        self.parent_stack.append(node)
        result = super().visit(node)
        # 完成当前节点的处理后，将其从栈中弹出
        self.parent_stack.pop()
        return result

    def find_call(self, node: ast.Attribute):
        ''' A().b.c() => [A(), b, c()] '''
        node_list = []
        while isinstance(node, ast.Attribute):
            node_list.append(node)
            node = node.value

        if isinstance(node, ast.Call):
            node_list.append(node)
            return node_list[::-1]
        else:
            return None

    def visit_Call(self, node):
        # 创建一个新的Call节点，调用dict函数并传递关键字参数
        call_node = ast.Call(
            func=ast.Name(id='dict', ctx=ast.Load()),
            args=[],
            keywords=[]
        )

        partial_flag = False

        # skip node that cannot transform
        parent_node = self.parent_stack[-2] if len(self.parent_stack) > 1 else None
        if not isinstance(parent_node, self.transform_parent):
            return node

        if isinstance(node.func, (ast.Attribute, ast.Call)):
            node_list = self.find_call(node.func)
            if node_list is None:
                call_node.keywords.append(
                    ast.keyword(arg='_target_', value=ast.Attribute(attr=node.func.attr, value=node.func.value, ctx=ast.Load())))
            else:
                prev_node = self.visit_Call(node_list[0])
                for node_attr in node_list[1:]:
                    prev_node = ast.Call(
                        func=ast.Name(id='dict', ctx=ast.Load()),
                        args=[],
                        keywords=[
                            ast.keyword(arg='_target_', value=ast.Name(id='getattr', ctx=ast.Load())),
                            ast.keyword(arg='_args_',
                                        value=ast.List(elts=[prev_node, ast.Constant(value=node_attr.attr)], ctx=ast.Load())),
                        ]
                    )
                call_node.keywords.append(
                    ast.keyword(arg='_target_', value=prev_node)
                )

        else:
            if node.func.id == 'partial':
                partial_flag = True
            elif node.func.id != 'dict':
                call_node.keywords.append(ast.keyword(arg='_target_', value=ast.Name(id=node.func.id, ctx=ast.Load())))

        # 处理位置参数
        if node.args:
            args_list = [self.visit(arg) for arg in node.args]
            if partial_flag:
                call_node.keywords.append(ast.keyword(arg='_target_', value=args_list[0]))
                call_node.keywords.append(ast.keyword(arg='_partial_', value=ast.NameConstant(value=True)))
                if len(args_list) > 1:
                    args_list = ast.List(elts=args_list[1:], ctx=ast.Load())
                    call_node.keywords.append(ast.keyword(arg='_args_', value=args_list))
            else:
                # 创建一个列表节点，包含所有的位置参数
                args_list = ast.List(elts=args_list, ctx=ast.Load())
                # 将列表节点作为字典的值
                call_node.keywords.append(ast.keyword(arg='_args_', value=args_list))

        # 处理关键字参数
        if node.keywords:
            for i, keyword in enumerate(node.keywords):
                if keyword.arg is None and isinstance(keyword.value, ast.Call):  # merge '**neko_cfg()' to dict
                    call_node.keywords.append(ast.keyword(arg=f'_merge_{i}_', value=self.visit(keyword.value)))
                else:
                    call_node.keywords.append(ast.keyword(arg=keyword.arg, value=self.visit(keyword.value)))

        # 替换原始的Call节点
        return call_node

    def visit_With(self, node):
        for item in node.items:
            if isinstance(item.context_expr, ast.Name) and item.context_expr.id == 'disable_neko_cfg':
                return node
        super().generic_visit(node)
        return node


class PythonCfgParser(YamlCfgParser):
    def __init__(self):
        super().__init__()
        self.cfg_dict = {}

    def print_code(self, code):
        # format code with yapf
        from yapf.yapflib.yapf_api import FormatCode
        new_code, _ = FormatCode(code, style_config='facebook')
        print(new_code)

    def compile_cfg(self, func):
        source = inspect.getsource(func)  # get source code
        tree = ast.parse(source)  # AST tree
        tree_func = tree.body[0]
        tree_func.decorator_list = [x for x in tree_func.decorator_list if x.id != 'neko_cfg']

        modified_tree = CallTransformer().visit(tree)
        ast.fix_missing_locations(modified_tree)

        # add source code for new function
        filename = random_filename()
        new_code = ast.unparse(modified_tree)
        lines = [line + '\n' for line in new_code.splitlines()]
        linecache.cache[filename] = (len(source), None, lines, filename)

        # compile modified code
        code = compile(modified_tree, filename=filename, mode="exec")
        namespace = func.__globals__
        exec(code, namespace)
        f_new = namespace[func.__name__]
        f_new._neko_cfg_ = True
        return f_new

    def resolve_sub_cfgs(self, cfg):
        if isinstance(cfg, dict):
            if '_target_' in cfg and getattr(cfg['_target_'], '_neko_cfg_', False):
                target = cfg.pop('_target_')
                if '_args_' in cfg:
                    args = cfg.pop('_args_')
                    cfg = target(*args, **cfg)
                else:
                    cfg = target(**cfg)
                res = self.resolve_sub_cfgs(cfg)
                return res or cfg

            cfg_new = {}
            for key, value in cfg.items():
                if isinstance(value, dict):
                    if re.match(r'_merge_\d+_', key):
                        res = self.resolve_sub_cfgs(value)
                        cfg_new.update(**res)
                    else:
                        res = self.resolve_sub_cfgs(value)
                        cfg_new[key] = res or value
                elif isinstance(value, list):
                    self.resolve_sub_cfgs(value)
                    cfg_new[key] = value
                else:
                    cfg_new[key] = value
            return cfg_new
        elif isinstance(cfg, list):
            for idx, value in enumerate(cfg):
                if isinstance(value, dict):
                    res = self.resolve_sub_cfgs(value)
                    if res is not None:
                        cfg[idx] = res
                elif isinstance(value, list):
                    self.resolve_sub_cfgs(value)

    def load_cfg(self, path: Union[str, ModuleType]):
        if isinstance(path, str):
            # record for save
            if len(self.cfg_dict) == 0:
                self.cfg_dict['cfg.py'] = path
            else:
                self.cfg_dict[path] = path

            # load module
            module_name = os.path.splitext(os.path.basename(path))[0]
            spec = importlib.util.spec_from_file_location(module_name, path)
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
        elif isinstance(path, ModuleType):
            # record for save
            cfg_path = get_rel_path(path)
            if len(self.cfg_dict) == 0:
                self.cfg_dict['cfg.py'] = cfg_path
            else:
                self.cfg_dict[cfg_path] = cfg_path

            module = path
        else:
            raise ValueError(f'base must be str or module. But type of base is {type(path)}')

        if getattr(module.make_cfg, '_neko_cfg_', False):
            cfg = module.make_cfg()
        else:
            f_cfg = self.compile_cfg(module.make_cfg)
            cfg = f_cfg()
        cfg = self.resolve_sub_cfgs(cfg) or cfg

        return OmegaConf.create(cfg, flags={"allow_objects": True})

    def save_configs(self, cfg, path, name='cfg'):
        for dst, src in self.cfg_dict.items():
            if dst == 'cfg.py':
                path_dst = os.path.join(path, f'{name}.py')
            else:
                path_dst = os.path.join(path, dst)
            os.makedirs(os.path.dirname(path_dst), exist_ok=True)
            shutil.copy2(src, path_dst)



import torch
from torch import nn
from torch.nn import functional as F
from einops import repeat

from .base import LossContainer


class MLCEImageLoss(LossContainer):
    def __init__(self, weight=1.0, eps=1e-4, reduction='mean'):
        super().__init__(None, weight=weight)
        self.ce_loss = nn.CrossEntropyLoss()
        self.eps = eps
        self.reduction = reduction

    def forward(self, pred, target):
        input_tensor = pred['pred']  # [B,C]
        target_tensor = target['label']

        if 'label_weight' in target:
            input_tensor = input_tensor / target['label_weight']  # softmax(w*x)

        log_prob_raw = F.softmax(input_tensor, dim=1)

        same_mask = (target_tensor.unsqueeze(0) == target_tensor.unsqueeze(1)).long()  # [B,B]
        same_mask_diag0 = same_mask - torch.diag_embed(torch.diag(same_mask))  # diag=0

        log_prob_x = log_prob_raw.clone()
        log_prob_x[same_mask_diag0.bool()] = self.eps
        log_prob_x.diagonal().copy_((log_prob_raw * same_mask_diag0.detach()).sum(dim=1))
        log_prob_x = log_prob_x + torch.diag_embed(torch.ones(len(target_tensor)) * self.eps).to(input_tensor.device)
        y = torch.arange(0, len(target_tensor)).to(input_tensor.device)

        return F.nll_loss(log_prob_x.log(), y, reduction=self.reduction) * self.alpha


# Copy from https://github.com/RElbers/info-nce-pytorch/blob/main/info_nce/__init__.py
class InfoNCELoss(LossContainer):
    """
    Calculates the InfoNCE loss for self-supervised learning.
    This contrastive loss enforces the embeddings of similar (positive) samples to be close
        and those of different (negative) samples to be distant.
    A query embedding is compared with one positive key and with one or more negative keys.

    References:
        https://arxiv.org/abs/1807.03748v2
        https://arxiv.org/abs/2010.05113

    Args:
        temperature: Logits are divided by temperature before calculating the cross entropy.
        reduction: Reduction method applied to the output.
            Value must be one of ['none', 'sum', 'mean'].
            See torch.nn.functional.cross_entropy for more details about each option.
        negative_mode: Determines how the (optional) negative_keys are handled.
            Value must be one of ['paired', 'unpaired'].
            If 'paired', then each query sample is paired with a number of negative keys.
            Comparable to a triplet loss, but with multiple negatives per sample.
            If 'unpaired', then the set of negative keys are all unrelated to any positive key.

    Input shape:
        query: (N, D) Tensor with query samples (e.g. embeddings of the input).
        positive_key: (N, D) Tensor with positive samples (e.g. embeddings of augmented input).
        negative_keys (optional): Tensor with negative samples (e.g. embeddings of other inputs)
            If negative_mode = 'paired', then negative_keys is a (N, M, D) Tensor.
            If negative_mode = 'unpaired', then negative_keys is a (M, D) Tensor.
            If None, then the negative keys for a sample are the positive keys for the other samples.

    Returns:
         Value of the InfoNCE Loss.

     Examples:
        >>> loss = InfoNCE()
        >>> batch_size, num_negative, embedding_size = 32, 48, 128
        >>> query = torch.randn(batch_size, embedding_size)
        >>> positive_key = torch.randn(batch_size, embedding_size)
        >>> negative_keys = torch.randn(num_negative, embedding_size)
        >>> output = loss(query, positive_key, negative_keys)
    """

    def __init__(self, weight=1.0, temperature=0.1, reduction='mean', negative_mode='unpaired'):
        super().__init__(None, weight=weight)
        self.temperature = temperature
        self.reduction = reduction
        self.negative_mode = negative_mode

    def forward(self, pred, target):
        input_tensor = pred['feat'][0]  # [B,C]
        target_tensor = target['label']

        if 'label_neg' in target:  # 显式给出pos和neg
            query = input_tensor[target_tensor == 0, :]
            positive_key = input_tensor[target_tensor == 1, :]
            negative_keys = input_tensor[target_tensor == 2, :]
        else:
            pos = target_tensor[0] == target_tensor
            query, positive_key = input_tensor[pos, :].chunk(2)
            if query.shape[0] != positive_key.shape[0]:
                query = query[:-1]
            negative_keys = input_tensor[~pos, :]

        return self.info_nce(query, positive_key, negative_keys, T=self.temperature) * self.alpha

    def info_nce(self, query, pos, neg, T=0.1):
        '''
        :param query: [Nq, C]
        :param pos: [Np, C]
        :param neg: [Nn, C]
        '''
        query = F.normalize(query, dim=-1)
        pos = F.normalize(pos, dim=-1)
        neg = F.normalize(neg, dim=-1)

        sim_pos = torch.mm(query, pos.transpose(0, 1))  # [Nq, Np]
        sim_neg = torch.mm(query, neg.transpose(0, 1))  # [Nq, Nn]
        logits = torch.cat([sim_pos, sim_neg], dim=1)
        labels = torch.cat([torch.ones(sim_pos.shape[-1]), torch.zeros(sim_neg.shape[-1])]).to(query.device,
                                                                                               dtype=query.dtype)
        labels = repeat(labels, 'n -> Nq n', Nq=len(query))
        return F.cross_entropy(logits / T, labels, reduction=self.reduction)


class NoisyInfoNCELoss(InfoNCELoss):
    def info_nce(self, query, pos, neg, T=0.1):
        '''
        :param query: [Nq, C]
        :param pos: [Np, C]
        :param neg: [Nn, C]
        '''
        query = F.normalize(query, dim=-1)
        pos = F.normalize(pos, dim=-1)
        neg = F.normalize(neg, dim=-1)

        sim_pos = torch.mm(query, pos.transpose(0, 1))  # [Nq, Np]
        sim_neg = torch.mm(query, neg.transpose(0, 1))  # [Nq, Nn]
        logits = torch.cat([sim_pos, sim_neg], dim=1)
        labels = torch.cat([torch.ones(sim_pos.shape[-1]), torch.zeros(sim_neg.shape[-1])]).to(query.device,
                                                                                               dtype=query.dtype)
        labels = repeat(labels, 'n -> Nq n', Nq=len(query))

        # RCE+NCE https://arxiv.org/pdf/2006.13554.pdf
        rce = self.reverse_cross_entropy(logits / T, labels)
        nce = self.normalized_cross_entropy(logits / T, labels)
        return rce + nce

    def reverse_cross_entropy(self, pred, labels):
        pred = F.softmax(pred, dim=1)
        pred = torch.clamp(pred, min=1e-7, max=1.0)
        labels = torch.clamp(labels, min=1e-4, max=1.0)
        rce = -torch.sum(pred * torch.log(labels), dim=1)
        return rce

    def normalized_cross_entropy(self, pred, labels):
        pred = F.log_softmax(pred, dim=1)
        nce = -torch.sum(labels * pred, dim=1) / (-pred.sum(dim=1))
        return nce



import torch
from torch import nn
from torch.nn import functional as F

from .base import LossContainer
from rainbowneko.models.layers import StyleSimilarity
import math


class StyleLoss(nn.Module):
    def __init__(self, eps=1e-4, reduction='mean'):
        super().__init__()
        self.ce_loss = nn.CrossEntropyLoss()
        self.eps = eps
        self.reduction = reduction
        self.style_sim = StyleSimilarity()

    _key_map = {'input_tensor': 'pred.feat', 'target_tensor': 'pred.label'}
    def forward(self, input_tensor, target_tensor):
        # input_tensor: list([B,C,H,W])

        if 2 in target_tensor:  # triplet style loss
            query = [item[target_tensor == 0, ...] for item in input_tensor]
            positive_key = [item[target_tensor == 1, ...] for item in input_tensor]
            negative_keys = [item[target_tensor == 2, ...] for item in input_tensor]

            return sum(self.style_sim(q, pos) + F.relu(-self.style_sim(q, neg)+5) for q, pos, neg in zip(query, positive_key, negative_keys))/len(query)
            
            #loss_pos = sum(self.style_sim(q, pos) for q, pos in zip(query, positive_key))/len(query)
            #loss_neg = sum(self.style_sim(q, neg) for q, neg in zip(query, negative_keys))/len(query)
            #return loss_pos + F.relu(-loss_neg+5)
        else:
            query = [item[target_tensor == 0, ...] for item in input_tensor]
            positive_key = [item[target_tensor == 1, ...] for item in input_tensor]
            return sum(self.style_sim(q, pos) for q, pos in zip(query, positive_key))/len(query)



from typing import Union, Iterable, Dict, Any, List

from addict import Dict as ADict

from .utils import is_dict, dict_parse_list, dict_list_copy, remove_empty_dict_list


class KeyMapper:
    cls_map = {0: 'pred.pred', 1: 'inputs.label'}  # 'pred.pred -> 0', 'inputs.label -> 1'
    image_map = {0: 'pred.pred', 1: 'inputs.target_image'}

    def __init__(self, host=None, key_map: Union[Iterable[str], Dict[Any, str], "KeyMapper"] = None, skip_missing=True,
                 move_mode=False):
        self.skip_missing = skip_missing
        self.move_mode = move_mode

        if key_map is None:
            if host is None:
                self.key_map = None
            else:
                if hasattr(host, '_key_map'):
                    self.key_map = self.parse_key_map(host._key_map)
                else:
                    self.key_map = self.cls_map
        elif isinstance(key_map, KeyMapper):
            self.key_map = key_map.key_map
            self.skip_missing = key_map.skip_missing
            self.move_mode = key_map.move_mode
        else:
            self.key_map = self.parse_key_map(key_map)

    def parse_key_map(self, key_map: Union[Iterable[str], Dict[Any, str]]):
        if is_dict(key_map):
            return key_map
        else:
            key_map_parse = {}
            for x in key_map:
                src, dst = x.split('->')
                src, dst = src.strip(), dst.strip()

                try:
                    src = int(src)
                except:
                    pass

                try:
                    dst = int(dst)
                except:
                    pass

                key_map_parse[dst] = src
            return key_map_parse

    def get_value(self, v, keys: Union[str, List]):
        if isinstance(keys, str):
            keys = keys.split('.')
        elif isinstance(keys, int):
            keys = [keys]

        for i, k in enumerate(keys):
            if isinstance(v, (list, tuple)):
                try:
                    k = int(k)
                except ValueError:
                    raise ValueError(f'{k} in {keys} cannot index a list')
                if self.move_mode and i == len(keys) - 1:
                    v = v.pop(k)
                else:
                    v = v[k]
            else:
                if self.move_mode and i == len(keys) - 1:
                    v = v.pop(k)
                else:
                    v = v[k]
        return v

    def set_value(self, v, keys: Union[str, List], data: ADict):
        if isinstance(keys, str):
            keys = keys.split('.')
        elif isinstance(keys, int):
            keys = [keys]

        try:
            k = int(keys[0])
            data = data['args']
        except ValueError:
            data = data['kwargs']

        for k in keys[:-1]:
            try:
                k = int(k)
            except ValueError:
                pass
            data = data[k]

        try:
            k = int(keys[-1])
            data[k] = v
        except ValueError:
            data[keys[-1]] = v

    def map_data(self, src):
        if self.key_map is None:
            return [], src

        if self.move_mode:
            src = dict_list_copy(src)

        data = ADict({'args': {}, 'kwargs': {}})
        for k_dst, k_src in self.key_map.items():
            if self.skip_missing:
                try:
                    v = self.get_value(src, k_src)
                    self.set_value(v, k_dst, data)
                except KeyError:
                    pass
                except IndexError:
                    pass
            else:
                v = self.get_value(src, k_src)
                self.set_value(v, k_dst, data)
        if self.move_mode:
            src_ = ADict(src)
            src_.update(data['kwargs'])
            src_ = remove_empty_dict_list(src_)
            data['kwargs'] = src_
        data = dict_parse_list(data.to_dict())
        args, kwargs = data['args'], data['kwargs']

        return args, kwargs

    def __call__(self, **src):
        return self.map_data(src)


if __name__ == '__main__':
    mapper = KeyMapper(key_map=['0 -> input', '1 -> target.label'])
    data = ('t1', 't2')

    args, kwargs = mapper.map_data(data)
    print(args, kwargs)

    mapper = KeyMapper(key_map=['image -> input', 'target.label -> 0', 'target.image -> 1'])
    data = {'image': 'image_data', 'target': {'label': 'label_data'}}

    args, kwargs = mapper.map_data(data)
    print(args, kwargs)

    mapper = KeyMapper(key_map=['target.label -> label_1'], move_mode=True)
    args, kwargs = mapper.map_data(data)
    print(args, kwargs)  # {} {'image': 'image_data', 'label_1': 'label_data'}

    mapper = KeyMapper(key_map=['target.label -> label_1'], move_mode=True)
    data = {'image': 'image_data', 'target': {'label': 'label_data', 'cls': 'cls_data'}}
    args, kwargs = mapper.map_data(data)
    print(args, kwargs)  # {} {'image': 'image_data', 'target': {'cls': 'cls_data'}, 'label_1': 'label_data'}



from functools import partial

import torch
import torchvision
from torch import nn
from torch.nn import CrossEntropyLoss

from cfgs.py.train.classify import multi_class
from rainbowneko.ckpt_manager import ckpt_saver
from rainbowneko.models.wrapper import DistillationWrapper
from rainbowneko.parser import CfgModelParser, neko_cfg
from rainbowneko.train.loss import LossContainer, LossGroup, DistillationLoss

num_classes = 10


def load_resnet(model, path=None):
    model.fc = nn.Linear(model.fc.in_features, num_classes)
    if path:
        model.load_state_dict(torch.load(path)['base'])
    return model

@neko_cfg
def make_cfg():
    return dict(
        _base_=[multi_class],

        model_part=CfgModelParser([
            dict(
                lr=1e-2,
                layers=['model_student'],
            )
        ]),

        ckpt_saver=dict(
            model=ckpt_saver(target_module='model_student')
        ),

        train=dict(
            train_epochs=100,
            save_step=2000,

            loss=LossGroup(_replace_=True, loss_list=[
                LossContainer(CrossEntropyLoss(), weight=0.05),
                LossContainer(DistillationLoss(T=5.0, weight=0.95)),
            ]),
        ),

        model=dict(
            name='cifar-resnet18',
            wrapper=partial(DistillationWrapper, _replace_=True,
                            model_teacher=load_resnet(torchvision.models.resnet50()),
                            model_student=load_resnet(torchvision.models.resnet18())
                            )
        ),
    )



import setuptools
import os
import platform

with open("README.md", "r", encoding='utf8') as fh:
    long_description = fh.read()

requires = []
with open('requirements.txt', encoding='utf8') as f:
    for x in f.readlines():
        requires.append(f'{x.strip()}')

if platform.system().lower() == 'windows':
    requires.append('bitsandbytes-windows')
else:
    requires.append('bitsandbytes')

def get_data_files(data_dir, prefix=''):
    file_dict = {}
    for root, dirs, files in os.walk(data_dir, topdown=False):
        for name in files:
            if prefix+root not in file_dict:
                file_dict[prefix+root] = []
            file_dict[prefix+root].append(os.path.join(root, name))
    return [(k, v) for k, v in file_dict.items()]


setuptools.setup(
    name="rainbowneko",
    py_modules=["rainbowneko"],
    version="1.5.1",
    author="IrisRainbowNeko",
    author_email="rainbow-neko@outlook.com",
    description="Neural network training and inference framework",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/IrisRainbowNeko/RainbowNekoEngine",
    packages=setuptools.find_packages(),
    classifiers=[
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
    ],
    python_requires='>=3.8',

    entry_points={
        'console_scripts': [
            'nekoinit = rainbowneko.tools.init_proj:main',
            'neko_train = rainbowneko.train.trainer.trainer_ac:neko_train',
            'neko_train_1gpu = rainbowneko.train.trainer.trainer_ac_single:neko_train',
            'neko_run = rainbowneko.infer.infer_workflow:run_workflow',
        ]
    },

    include_package_data=True,

    install_requires=requires
)



import cv2
import numpy as np
from PIL import Image
from torchvision import transforms as T
from torchvision.transforms import functional as F

class DualRandomCrop:
    def __init__(self, size):
        self.size = size

    def __call__(self, img):
        crop_params = T.RandomCrop.get_params(img['image'], self.size)
        img['image'] = F.crop(img['image'], *crop_params)
        if "mask" in img:
            img['mask'] = self.crop(img['mask'], *crop_params)
        if "cond" in img:
            img['cond'] = F.crop(img['cond'], *crop_params)
        return img, crop_params[:2]

    @staticmethod
    def crop(img: np.ndarray, top: int, left: int, height: int, width: int) -> np.ndarray:
        right = left+width
        bottom = top+height
        return img[top:bottom, left:right, ...]

def resize_crop_fix(img, target_size, mask_interp=cv2.INTER_CUBIC):
    w, h = img['image'].size
    if w == target_size[0] and h == target_size[1]:
        return img, [h,w,0,0,h,w]

    ratio_img = w/h
    if ratio_img>target_size[0]/target_size[1]:
        new_size = (round(ratio_img*target_size[1]), target_size[1])
        interp_type = Image.LANCZOS
    else:
        new_size = (target_size[0], round(target_size[0]/ratio_img))
        interp_type = Image.LANCZOS
    img['image'] = img['image'].resize(new_size, interp_type)
    if "mask" in img:
        img['mask'] = cv2.resize(img['mask'], new_size, interpolation=mask_interp)
    if "cond" in img:
        img['cond'] = img['cond'].resize(new_size, interp_type)

    img, crop_coord = DualRandomCrop(target_size[::-1])(img)
    return img, [*new_size, *crop_coord[::-1], *target_size]

def pad_crop_fix(img, target_size):
    w, h = img['image'].size
    if w == target_size[0] and h == target_size[1]:
        return img, (h,w,0,0,h,w)

    pad_size = [0, 0, max(target_size[0]-w, 0), max(target_size[1]-h, 0)]
    if pad_size[2]>0 or pad_size[3]>0:
        img['image'] = F.pad(img['image'], pad_size)
        if "mask" in img:
            img['mask'] = np.pad(img['mask'], ((0, pad_size[3]), (0, pad_size[2])), 'constant', constant_values=(0, 0))
        if "cond" in img:
            img['cond'] = F.pad(img['cond'], pad_size)

    if pad_size[2]>0 and pad_size[3]>0:
        return img, (h,w,0,0,h,w)  # No need to crop
    else:
        img, crop_coord = DualRandomCrop(target_size[::-1])(img)
        return img, crop_coord

class CycleData():
    def __init__(self, data_loader):
        self.data_loader = data_loader

    def __iter__(self):
        self.epoch = 0

        def cycle():
            while True:
                self.data_loader.dataset.bucket.rest(self.epoch)
                for data in self.data_loader:
                    yield data
                self.epoch += 1

        return cycle()



from typing import Dict

from .dataset import BaseDataset, BaseBucket, DataHandler
from .source import WebDatasetSource


class WebDataset(BaseDataset):
    def __init__(self, bucket: BaseBucket = None, source: Dict[str, WebDatasetSource] = None, handler: DataHandler = None,
                 batch_handler: DataHandler = None, size: int = 10000, shuffle=True, **kwargs):
        assert all(isinstance(source, WebDatasetSource) for source in source.values()), 'WebDataset only accept WebDatasetSource.'
        self.size = size
        self.shuffle = shuffle
        super().__init__(bucket=bucket, source=source, handler=handler, batch_handler=batch_handler, **kwargs)

    def __len__(self):
        return self.size

    def __getitem__(self, index):
        datas = self.bucket.next_data(shuffle=self.shuffle)
        datas = self.handler(datas)
        return datas



from functools import partial
from typing import List

from tqdm.auto import tqdm

from rainbowneko.utils import KeyMapper, dict_merge


class BasicAction:
    feedback_input = True

    def __init__(self, key_map_in=None, key_map_out=None):
        self.key_mapper_in = KeyMapper(key_map=key_map_in)
        self.key_mapper_out = KeyMapper(key_map=key_map_out)

    def __call__(self, **states):
        _, inputs = self.key_mapper_in.map_data(states)
        output = self.forward(**inputs)
        if output is not None:
            _, output = self.key_mapper_out.map_data(output)
            if self.feedback_input:
                states = dict_merge(states, output)
            else:
                return output
        return states

    def forward(self, **states):
        raise NotImplementedError()


class FromMemory:
    def __init__(self, action: partial[BasicAction], key_map_in=None):
        self.action = action
        self.key_mapper_in = KeyMapper(key_map=key_map_in)

    def __call__(self, **states):
        _, inputs = self.key_mapper_in.map_data(states)
        action = self.action(**inputs)

        return action(**states)


class Actions(BasicAction):
    feedback_input = False

    def __init__(self, actions: List[BasicAction], key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.actions = actions

    def forward(self, **states):
        pbar = tqdm(self.actions, leave=False)
        N_steps = len(self.actions)
        for step, act in enumerate(pbar):
            pbar.set_description(f'[{step + 1}/{N_steps}] action: {type(act).__name__}')
            states = act(**states)
            # print(f'states: {", ".join(states.keys())}')
        return states


class LoopAction(BasicAction):
    feedback_input = False

    def __init__(self, iterator, actions: List[BasicAction]):
        self.iterator = iterator
        self.actions = actions

    def __call__(self, **states):
        iterator = self.iterator(**states)

        pbar = tqdm(iterator)
        N_steps = len(self.actions)
        for data in pbar:
            states.update(data)
            for step, act in enumerate(self.actions):
                pbar.set_description(f'[{step + 1}/{N_steps}] action: {type(act).__name__}')
                states = act(**states)
        return states


class IterAction(BasicAction):
    feedback_input = False

    def __init__(self, actions: List[BasicAction], **loop_value):
        self.loop_value = loop_value
        self.actions = actions

    def forward(self, **states):
        loop_data = [states.pop(k) for k in self.loop_value.keys()]
        pbar = tqdm(zip(*loop_data), total=len(loop_data[0]))
        N_steps = len(self.actions)
        for data in pbar:
            feed_data = {k: v for k, v in zip(self.loop_value.values(), data)}
            states.update(feed_data)
            for step, act in enumerate(self.actions):
                pbar.set_description(f'[{step + 1}/{N_steps}] action: {type(act).__name__}')
                states = act(**states)
        return states


class LambdaAction(BasicAction):
    def __init__(self, f_act, key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.f_act = f_act

    def forward(self, **states):
        out = self.f_act(**states)
        return out



from .pkl import PKLFormat
from .safetensor import SafeTensorFormat
from .base import CkptFormat

try:
    from .onnx import ONNXFormat
except:
    print('ONNXFormat not available, onnx not installed.')


from typing import Iterable, Any, Dict

from omegaconf import OmegaConf, ListConfig, DictConfig
import os


class YamlCfgParser:

    def load_cfg(self, path: str):
        return OmegaConf.load(path)

    def merge_configs(self, source: Dict, overrides: Dict):
        self.remove_replace(source, overrides)
        return OmegaConf.merge(source, overrides)

    def remove_config_undefined(self, cfg):
        itr: Iterable[Any] = range(len(cfg)) if isinstance(cfg, ListConfig) else cfg

        undefined_keys = []
        for key in itr:
            if cfg._get_child(key) == '---':
                undefined_keys.append(key)
            elif OmegaConf.is_config(cfg[key]):
                self.remove_config_undefined(cfg[key])
        for key in undefined_keys:
            del cfg[key]
        return cfg

    def remove_replace(self, base_cfg, cfg):
        if isinstance(cfg, ListConfig) or base_cfg is None:
            return base_cfg

        replace_keys = []
        for key in cfg:
            child = cfg._get_child(key)
            if isinstance(child, DictConfig):
                if child._get_child('_replace_'):
                    if key in base_cfg:
                        replace_keys.append(key)
                    continue
                else:
                    if key in base_cfg:
                        self.remove_replace(base_cfg[key], child)
                    else:
                        continue
        for key in replace_keys:
            del base_cfg[key]
            del cfg[key]['_replace_']
        return base_cfg

    def load_config(self, path, remove_undefined=True):
        cfg = self.load_cfg(path)
        if '_base_' in cfg:
            for base in cfg['_base_']:
                cfg = self.merge_configs(self.load_config(base, remove_undefined=False), cfg)
            del cfg['_base_']
        if remove_undefined:
            cfg = self.remove_config_undefined(cfg)
        return cfg

    def load_config_with_cli(self, path, args_list=None, remove_undefined=True):
        cfg = self.load_config(path, remove_undefined=False)
        cfg_cli = OmegaConf.from_cli(args_list)
        cfg = OmegaConf.merge(cfg, cfg_cli)
        if remove_undefined:
            cfg = self.remove_config_undefined(cfg)
        return cfg

    def save_configs(self, cfg, path, name='cfg'):
        OmegaConf.save(cfg, os.path.join(path, f'{name}.yaml'))


from .base import DataSource, ComposeDataSource, VisionDataSource
from .index import IndexSource
from .img_label import ImageLabelSource
from .img_pair import ImagePairSource
from .folder_class import ImageFolderClassSource
from .unlabel import UnLabelSource

try:
    from .webds import WebDatasetSource, WebDSImageLabelSource
except (ImportError, ModuleNotFoundError):
    print('INFO: webdataset is not available')


import os.path

import torch
from tqdm.auto import tqdm

from pathlib import Path
from rainbowneko.models.wrapper import BaseWrapper
from torch.utils.data import DataLoader
from .evaluator import Evaluator, WorkflowEvaluator
from .renderer import Renderer


class Previewer(Evaluator):
    def __init__(self, trainer: "Trainer", data_loader: DataLoader, img_dir: str, renderer: Renderer, ds_name=None, interval=100, dataset:None=None):
        assert dataset is None, '"dataset" is a placeholder for cfg, should always be None.'
        self.data_loader = data_loader
        self.trainer = trainer
        self.img_dir = img_dir
        self.renderer = renderer
        self.interval = interval
        self.ds_name = ds_name

    def save_images(self, pred, target, img_dir, step):
        img_list = self.renderer(pred, target)
        for img in img_list:
            img.save(os.path.join(img_dir, f"{step}-{self.img_count}.png"))
            self.img_count += 1

    @torch.no_grad()
    def evaluate(self, step: int, model: BaseWrapper):
        if step % self.interval != 0:
            return

        # record training layers
        training_layers = [layer for layer in model.modules() if layer.training]

        model.eval()

        img_dir = os.path.join(self.trainer.exp_dir, self.img_dir)
        self.trainer.loggers.info(f'Preview to {img_dir}')
        self.img_count=0
        for data in tqdm(self.data_loader, disable=not self.trainer.is_local_main_process):
            pred, input_datas = self.forward_one_step(model, data)
            self.save_images(pred, input_datas, img_dir=img_dir, step=step)

        for layer in training_layers:
            layer.train()

    def to(self, device):
        self.renderer.to(device)


class WorkflowPreviewer(WorkflowEvaluator):
    @torch.no_grad()
    def evaluate(self, step: int, model: BaseWrapper, prefix='eval/'):
        if step % self.interval != 0:
            return

        # record training layers
        training_layers = [layer for layer in model.modules() if layer.training]

        model.eval()
        self.trainer.loggers.info(f'Preview')

        states = self.workflow_runner.run(model=model, in_preview=True, device=self.device, dtype=self.dtype,
                                          preview_root=Path(self.trainer.exp_dir)/'imgs',
                                          world_size=self.trainer.world_size, local_rank=self.trainer.local_rank)

        for layer in training_layers:
            layer.train()


import torch
from torch import nn
import math

class GroupLinear(nn.Module):
    def __init__(self, in_features: int, out_features: int, group: int, bias: bool = True,
                 device=None, dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.group = group
        self.weight = nn.Parameter(torch.empty((group, out_features, in_features), **factory_kwargs))
        if bias:
            self.bias = nn.Parameter(torch.empty(group, 1, out_features, **factory_kwargs))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    @staticmethod
    def _calculate_fan_in_and_fan_out(tensor):
        dimensions = tensor.dim()
        if dimensions < 2:
            raise ValueError("Fan in and fan out can not be computed for tensor with fewer than 2 dimensions")

        num_input_fmaps = tensor.size(-2)
        num_output_fmaps = tensor.size(-1)
        fan_in = num_input_fmaps
        fan_out = num_output_fmaps

        return fan_in, fan_out

    @staticmethod
    def kaiming_uniform_(
            tensor: torch.Tensor,
            a: float = 0,
            nonlinearity: str = "leaky_relu",
            generator = None,
    ):
        fan_in, _ = GroupLinear._calculate_fan_in_and_fan_out(tensor)
        gain = nn.init.calculate_gain(nonlinearity, a)
        std = gain / math.sqrt(fan_in)
        bound = math.sqrt(3.0) * std  # Calculate uniform bounds from standard deviation
        with torch.no_grad():
            return tensor.uniform_(-bound, bound, generator=generator)


    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(in_features), 1/sqrt(in_features)). For details, see
        # https://github.com/pytorch/pytorch/issues/57109
        self.kaiming_uniform_(self.weight, a=math.sqrt(5))
        if self.bias is not None:
            fan_in, _ = self._calculate_fan_in_and_fan_out(self.weight)
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            nn.init.uniform_(self.bias, -bound, bound)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        '''
        input(G,N,CI) x weight.T(G,CI,CO) [+ bias(G,1,CO)] -> out(G,N,CO)
        '''
        if self.bias is None:
            return torch.bmm(input, self.weight.transpose(1,2))
        else:
            return torch.baddbmm(self.bias, input, self.weight.transpose(1, 2))

    def extra_repr(self) -> str:
        return f'in_features={self.in_features}, out_features={self.out_features}, group={self.group}, bias={self.bias is not None}'


from typing import Dict, Any, List
from rainbowneko.utils import KeyMapper, RandomContext

class DataHandler:
    def __init__(self, key_map_in=('image -> image',), key_map_out=('image -> image',)):
        self.key_mapper_in = KeyMapper(key_map=key_map_in)
        self.key_mapper_out = KeyMapper(key_map=key_map_out)

    def handle(self, *args, **kwargs):
        return kwargs

    def __call__(self, data) -> Dict[str, Any]:
        data_proc = self.handle(**self.key_mapper_in.map_data(data)[1])
        return self.key_mapper_out.map_data(data_proc)[1]

class HandlerGroup(DataHandler):
    def __init__(self, key_map_in=None, key_map_out=None, **handlers: DataHandler):
        self.handlers = handlers
        self.key_mapper_in = KeyMapper(key_map=key_map_in) if key_map_in else None
        self.key_mapper_out = KeyMapper(key_map=key_map_out) if key_map_out else None

    def __call__(self, data:Dict[str, Any]) -> Dict[str, Any]:
        if self.key_mapper_in:
            data = self.key_mapper_in.map_data(data)[1]
        data_new = {}
        for name, handler in self.handlers.items():
            data_new.update(handler(data))
        data.update(data_new)
        if self.key_mapper_out:
            data = self.key_mapper_out.map_data(data)[1]
        return data

class HandlerChain(DataHandler):
    def __init__(self, key_map_in=None, key_map_out=None, **handlers: DataHandler):
        self.handlers = handlers
        self.key_mapper_in = KeyMapper(key_map=key_map_in) if key_map_in else None
        self.key_mapper_out = KeyMapper(key_map=key_map_out) if key_map_out else None

    def __call__(self, data:Dict[str, Any]) -> Dict[str, Any]:
        data = data.copy()
        if self.key_mapper_in:
            data = self.key_mapper_in.map_data(data)[1]
        for name, handler in self.handlers.items():
            tmp = handler(data)
            data.update(tmp)
        if self.key_mapper_out:
            data = self.key_mapper_out.map_data(data)[1]
        return data

class SyncHandler(DataHandler):
    def __init__(self, **handlers: DataHandler):
        self.handlers = handlers

    def __call__(self, data) -> Dict[str, Any]:
        random_context = RandomContext()

        data_new = {}
        for name, handler in self.handlers.items():
            with random_context:
                data_new.update(handler(data))
        data.update(data_new)
        return data


from typing import List, Tuple, Union

from rainbowneko.utils import RandomContext

from .base import DataHandler
import torch


class FrameHandler(DataHandler):
    def __init__(self, handler: DataHandler, cat=False, key_map_in=('frames -> frames',), key_map_out=('frames -> frames',)):
        super().__init__(key_map_in, key_map_out)
        self.handler = handler
        self.cat = cat

    def handle(self, frames: Union[List, Tuple]):
        assert isinstance(frames, list) or isinstance(frames, tuple), "FrameHandler: frames should be a list or tuple"

        random_context = RandomContext()
        frames_new = []
        for frame in frames:
            with random_context:
                frames_new.append(self.handler(frame))
        if self.cat:
            frames_new = torch.cat(frames_new, 0)
        return {'frames': frames_new}


class FrameCatHandler(DataHandler):
    def __init__(self, key_map_in=('frames -> frames',), key_map_out=('frames -> frames',)):
        super().__init__(key_map_in, key_map_out)

    def handle(self, frames: Union[List, Tuple]):
        return {'frames': torch.stack(frames, 0)} # [T, ...]


from typing import Dict, Any, List, Union
import torch

from PIL import Image
from .base import BasicAction
from rainbowneko.ckpt_manager import NekoLoader

class LoadImageAction(BasicAction):
    def __init__(self, image_paths:Union[str, List[str]], image_transforms=None, key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        if isinstance(image_paths, str):
            image_paths = [image_paths]
        self.image_paths = image_paths
        self.image_transforms = image_transforms

    def load_one(self, path:str):
        img = Image.open(path).convert('RGB')
        if self.image_transforms:
            img = self.image_transforms(img)
        return img

    def forward(self, device, **states):
        input = torch.stack([self.load_one(path) for path in self.image_paths]).to(device)
        input: Dict[str, Any] = {'x':input}
        return {'input':input}

class LoadModelAction(BasicAction):
    def __init__(self, cfg: Dict[str, Any], key_map_in=None, key_map_out=None):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.cfg=cfg

    def forward(self, model, in_preview=False, **states):
        if not in_preview:
            NekoLoader.load_all(model, self.cfg)

class FeedAction(BasicAction):
    def __init__(self, key_map_in=None, key_map_out=None, **datas):
        super().__init__(key_map_in=key_map_in, key_map_out=key_map_out)
        self.datas = datas

    def forward(self, **states):
        return self.datas


from rainbowneko.utils import KeyMapper
from PIL import Image
from typing import List, Callable


class Renderer:
    def __init__(self, trans: Callable, key_map=KeyMapper.cls_map):
        self.trans = trans
        self.key_mapper = KeyMapper(None, key_map)

    def render(self, image, label) -> List[Image.Image]:
        return [self.trans(item) for item in image]

    def __call__(self, pred, target):
        args, kwargs = self.key_mapper(pred=pred, target=target)
        return self.render(*args, **kwargs)

    def to(self, device):
        pass


from .base import LossContainer, LossGroup
from .distillation import DistillationLoss
from .contrastive import MLCEImageLoss, InfoNCELoss, NoisyInfoNCELoss
from .style import StyleLoss
from .classify import SoftCELoss
from .pseudo_label import PseudoLabelLoss
from .charbonnier import CharbonnierLoss
from .ms_ssim import MSSSIMLoss